import numpy as np
import ivisit.matplotlib as ivml

# ***********************************************************************************
# Useful auxiliary functions 
# *************************************************************************
def getNearestNeighbor(X,x):
    """
    simple nearest neighbor search; returns index of nearest neighbor of x in X[1],X[2],...
    """
    dist = [np.linalg.norm(xi-x) for xi in X] # list of distances between x and X[i] 
    return np.argmin(dist)                    # return index idxNN of nearest neighbor X[idxNN] 

def computeGrid2D(f,xrange,yrange=None,f_mode='matrix'):
    """
    evaluate function f on 2D grid defined by xrange and yrange having format [min,max,delta]
    if f_mode is 'matrix' then f can process f(X,Y) for matrixes X,Y
    if f_mode is 'vector' then f can process f(X.flat,Y.flat) for matrixes X,Y
    if f_mode is 'scalar' then f can only process individual values f(x,y) for scalar values x,y
    returns X,Y,Z as matrixes each having the size of the grid
    """
    xmin,xmax,dx = xrange[0],xrange[1],xrange[2]
    if yrange is None: ymin,ymax,dy=xmin,xmax,dx
    else: ymin,ymax,dy = yrange[0],yrange[1],yrange[2]
    arange_x, arange_y = np.arange(xmin,xmax,dx), np.arange(ymin,ymax,dy)
    X,Y = np.meshgrid(arange_x, arange_y)
    if f_mode=='scalar':
        Z=np.array([[f([xj,yi]) for xj in arange_x] for yi in arange_y])
    elif f_mode=='vector':
        Z=f(X.flat,Y.flat)
        Z=Z.reshape(X.shape)
    elif f_mode=='matrix':
        Z=f(X,Y)
    return X,Y,Z

# ***********************************************************************************
# SimpleScope: a simple/fast oscilloscope-like class to display signals in x/y plots
#              independent of Matplotlib
# *************************************************************************
class SimpleScope:        # simple/fast oscilloscope-like class to display signals in x/y plots
    def __init__(self,im_width=100,im_height=100,min_x=0,max_x=100, min_y=0, max_y=100,
                 flagRGB=0, flag_line_x0=1, flag_line_y0=1,col_bkgr=0,col_frgr=255):
        self.configure(im_width,im_height,min_x,max_x,min_y,max_y,flagRGB,flag_line_x0,flag_line_y0,col_bkgr,col_frgr)

    def configure(self,im_width=None, im_height=None, min_x=None,max_x=None,min_y=None,max_y=None,
                  flagRGB=None,flag_line_x0=None,flag_line_y0=None,col_bkgr=None,col_frgr=None):
        # (i) set new parameters
        if not im_width     is None: self.im_width    =im_width
        if not im_height    is None: self.im_height   =im_height
        if not min_x        is None: self.min_x       =min_x
        if not max_x        is None: self.max_x       =max_x
        if not min_y        is None: self.min_y       =min_y
        if not max_y        is None: self.max_y       =max_y
        if not flagRGB      is None: self.flagRGB     =flagRGB
        if not flag_line_x0 is None: self.flag_line_x0=flag_line_x0
        if not flag_line_y0 is None: self.flag_line_y0=flag_line_y0
        if not col_bkgr     is None: self.col_bkgr    =col_bkgr
        if not col_frgr     is None: self.col_frgr    =col_frgr
        
        # (ii) allocate new image matrix im_data for data?
        if not im_width  is None or not im_height is None or not flagRGB is None:
            if self.flagRGB<=0: self.im_data=np.zeros((self.im_height,self.im_width)  ,'uint8') # gray image
            else              : self.im_data=np.zeros((self.im_height,self.im_width,3),'uint8') # RGB color image
            self.im_data[:,:]=self.col_bkgr    # initialize with background color

        # (iii) set transformation coefficients
        self.dx_inv_width  = self.im_width /(self.max_x-self.min_x)
        self.dy_inv_height = self.im_height/(self.max_y-self.min_y)
        self.i0=self.im_height-1-int((0-self.min_y)*self.dy_inv_height+0.5)%self.im_height # row index for y=0 line
        if self.flag_line_y0>0: self.im_data[self.i0,:]=self.col_frgr     # set y=0 line ?
        self.last_j,self.last_jp1=-1,0                                    # set dummy value for last_j (= last time position) and last_jp1 (=last x=0 line)

        return self.im_data   # return data image; assign this to ivisit data array

    def set_data(self,x,y_from,y_to=None,col=None,erase_x=1):  # draw at current time x point at x,y with color col (y,col may be lists)
        if col is None: col=self.col_frgr                           # default color is foreground
        j=int((x-self.min_x)*self.dx_inv_width+0.5)%self.im_width   # column/time index for data point
        jp1=(j+1)%self.im_width                                     # one column right of j (for new x=0 line)
        if erase_x>0 and j!=self.last_j: self.im_data[:,j]=self.col_bkgr  # erase column j ? (do not erase if j hasn't changed!)
        if self.flag_line_x0>0 and j!=self.last_j and j!=self.last_jp1: self.im_data[:,self.last_jp1]=self.col_bkgr  # erase old x=0 line if new j has leaped over 
        if self.flag_line_x0>0: self.im_data[:,jp1]=self.col_frgr   # draw line for x=0 ?
        if self.flag_line_y0>0: self.im_data[self.i0,j]=self.col_frgr     # restore y0=0 point
        if not isinstance(y_from,(list,tuple)): y_from,y_to,col=[y_from],[y_to],[col]  # cast y_from,y_to, and col as lists
        for ii in range(len(y_from)):
            i_from=self.im_height-1-int((y_from[ii]-self.min_y)*self.dy_inv_height+0.5)%self.im_height   # row index for data point
            if y_to is None: i_to=i_from
            else: i_to=self.im_height-1-int((y_to[ii]-self.min_y)*self.dy_inv_height+0.5)%self.im_height # row index for data point
            if i_from>i_to: i_to,i_from=i_from,i_to               # swap
            self.im_data[i_from:i_to+1,j]=col[ii]                 # set data point
        self.last_j=j                                             # keep last j (=last time position)
        self.last_jp1=jp1                                         # keep last j+1 (=position where the x=0 line has been drawn and should be deleted)


# *****************************************************************************************
# ClickDragEventAutomaton: Event handler automaton that can handle click and drag actions 
# *****************************************************************************************
class ClickDragEventAutomaton:
    def __init__(self, sim, var_action_name=None, handle_action=None, dict_action_type=None, button_no=1, flagMatplotlib=0):
        """
        :param sim: reference to IVISIT simulation object 
        :param var_action_name: name of the action variable defining the current action; must be attribute of self.sim.parameters
        :param handle_action: this function will be called to handle actions after click/drag events
                              handle_action should be a function of the simulation class (self.sim) and have the signature
                              def handle_action(self.sim,action,pos,pos_init   
        :param dict_action2type: dict defining for each action name (given by var_action_name) its type: either 'click' or 'drag' action
                                 example: dict_action2type={'New':'click', 'Delete':'click', 'Move':'drag'} defines that 
                                          actions "New" and "Delete" are click-actions (reuquire only one click), whereas
                                          action "Move" is a drag-action (requires initial-click, moving, releasing)
                                 if None or 'click' then all actions are interpreted as simple "click"-actions!
                                 if 'drag' then all actions are interpreted as drag actions!
        :param button_no: button number (either 1=left,2=middle,3=right button)
        :param flagMatplotlib: if >0 then click positions are transformed into Matplotlib coordinates (for clicking on Matplotlib figures!)
        """
        self.sim              = sim
        self.var_action_name  = var_action_name
        self.handle_action    = handle_action
        self.button_no        = str(button_no)
        self.dict_action_type = dict_action_type
        self.flagMatplotlib   = flagMatplotlib
        self.state            = 'IDLE'  # initialize state variable with IDLE (state can be either 'IDLE' or 'DRAG')
        self.pos, self.pos_init, self.pos_prev = None, None, None  # current, intial, and previous positions (of clicks or moves)
        self.action, self.action_type = None, None  # action (either "click", "drag_init", "drag_move", or "drag_finish") and action type (either "click" or "drag")

    def bind(self, display, img_var_name, img_widget_name, img_axis_name=None, img_widget_canvas_name="imgcanvas"):
        """
        :param display: display (usually a IVisitRawDisplay object) holding widgets to bind to
        :param img_var_name: name of the image matrix variable (typically a numpy array); must be attribute of self.sim.data
        :param img_widget_name: name of the IVISIT IMAGE widget where the the image is displayed (typically a TIVisitImageWidget)
        :param img_axis_name: name of the Matplotlib axis where the data points are drawn into (only for flagMatplotlib>0)
        :param img_widget_canvas_name: name of the canvas of the IVISIT IMAGE widget (default "imgcanvas" for TIVisitImageWidget) 
        """
        self.display                = display
        self.img_var_name           = img_var_name
        self.img_widget_name        = img_widget_name
        self.img_axis_name          = img_axis_name
        self.img_widget_canvas_name = img_widget_canvas_name
        if not display is None:
            display.bind2Widget(img_widget_name,"<Button-"+self.button_no+">"       ,self.onPressedButton ,img_widget_canvas_name)
            display.bind2Widget(img_widget_name,"<B"+self.button_no+"-Motion>"      ,self.onMovedButton   ,img_widget_canvas_name)
            display.bind2Widget(img_widget_name,"<ButtonRelease-"+self.button_no+">",self.onReleasedButton,img_widget_canvas_name)

    def getImgPos(self,event,flagCont=0,pad=1):
        """
        transform canvas positions event.x/y into image positions
        :param event: tkinter event with canvas positions event.x, event.y
        :param flagCont: if >0 then use continuous positions (e.g., to get Matplotlib positions)
        :param pad: tkinter canvas is usually larger than image matrix by pad in each direction (left, right, top, bottom)
        """
        canv = getattr(self.display.getWidget(self.img_widget_name),self.img_widget_canvas_name) # canvas of widget where mouse click ocurred
        canv_w,canv_h = canv.winfo_width()-2*pad, canv.winfo_height()-2*pad   # canvas size (without padding)
        img = getattr(self.sim.data,self.img_var_name)            # reference to image object
        x,y=max(0,(event.x-pad+0.5)*img.shape[1]/canv_w),max(0,(event.y-pad+0.5)*img.shape[0]/canv_h)  # continuous image position
        if flagCont<=0:          # discretize to image array indexes
            x,y=min(img.shape[1]-1,int(x)),min(img.shape[0]-1,int(y))
        return x,y
        
    def onPressedButton(self,event):  # called when clicking mouse button (button_no) on data widget
        if self.state=='IDLE':        # click on IMAGE only relevant in state IDLE
            self.sim.lock.acquire()   # make mouse processing thread-safe
            p,d=self.sim.parameters,self.sim.data # short hands to simulation parameters and data
            x,y=self.getImgPos(event,self.flagMatplotlib) # get image pixel position (note that canvas may be scaled compared to image)
            if self.flagMatplotlib>0: # get clicked position in Matplotlib coordinates?
                x,y=ivml.getDataPos_from_PixelPos(x,y,getattr(d,self.img_var_name).shape[0],getattr(self.sim,self.img_axis_name)) # Yes!
            self.pos=[x,y]            # store clicked position (in either image or Matplotlib coordinates)
            if self.dict_action_type is None: self.action_type="click"                       # per default action-type is click-action
            elif self.dict_action_type=='drag': self.action_type="drag"                      # drag action?
            else: self.action_type = self.dict_action_type[getattr(p,self.var_action_name)]  # get action-type from dict!
            if self.action_type=='click':
                self.action='click'        # if type is "click" then only action "click"
            elif self.action_type=='drag':                       
                self.action='drag_init'    # if type is "drag" then start drag sequence
                self.pos_init=[x,y]        # initial position 
                self.pos_prev=[x,y]        # previous position
                self.state='DRAG'          # change state to 'DRAG'
            else:
                assert 0, "unknown action_type="+str(self.action_type)+" (expected either 'click' or 'drag'!)"
            getattr(self.sim,self.handle_action)(self.action, self.pos, self.pos_init, self.pos_prev) # call to handle_action in simulation object
            self.sim.lock.release()   # release lock
        
    def onMovedButton(self,event):    # called when moved mouse while pressing button on data widget
        if self.state=='DRAG':        # moving mouse on IMAGE only relevant if in state DRAG
            self.sim.lock.acquire()   # make processing of mouse events thread-safe
            d=self.sim.data           # short hands to simulation parameters and data
            x,y=self.getImgPos(event,self.flagMatplotlib) # get image pixel position (note that canvas may be scaled compared to image)
            if self.flagMatplotlib>0: # get clicked position in Matplotlib coordinates?
                x,y=ivml.getDataPos_from_PixelPos(x,y,getattr(d,self.img_var_name).shape[0],getattr(self.sim,self.img_axis_name)) # Yes!
            self.pos_prev=self.pos    # previous position
            self.pos=[x,y]            # store clicked position
            self.action='drag_move'   # action 'drag_move' means the mouse has been moved to new position
            getattr(self.sim,self.handle_action)(self.action, self.pos, self.pos_init, self.pos_prev)  # call to handle_action in simulation object
            self.sim.lock.release()   # release lock
        
    def onReleasedButton(self,event): # called when releasing mouse button on data widget
        if self.state=='DRAG':        # releasing mouse on IMAGE only relevant if in state DRAG
            self.sim.lock.acquire()   # make processing of mouse events thread-safe
            self.action='drag_finish' # action 'drag_finish' means that mouse has been released and action should be finished
            getattr(self.sim,self.handle_action)(self.action, self.pos, self.pos_init, self.pos_prev) # call to handle_action in simulation object
            self.state='IDLE'         # change state back to 'IDLE'
            self.sim.lock.release()   # release lock
       



#!/usr/bin/python
# -*- coding: utf-8-unix -*-
    
################################################################################
# IViSiT 0.71: Interactive Visual Simulation Tool
#              based on Python/Tkinter for optimizing model parameters 
# progammed from September 2017 to July 2024 by Andreas Knoblauch
# All rights reserved
#
# Uses the tkinter widgets to implement a generic simulatiom framework
# including input/output widgets for online modifying of simulation parameters
# and online displaying of simulation states
#
# basic simulation framework inspired from neural simulation software PyFelix++
#
################################################################################
     
Version = '0.71'
__version__ = Version
import sys, os, threading, shutil, traceback           # platform, args, locks, run tools
import time,datetime
import __main__ as main
#import csv
#import numpy as np
#import matplotlib.dates
#from LevelMacherFrame import *
#import matplotlib.pyplot as plt    
from tkinter        import *               # base widgets, constants
from tkinter.filedialog import Open, SaveAs    # standard dialogs
from tkinter.messagebox import showinfo, showerror, askyesno
#from tkSimpleDialog import askstring, askinteger
#from tkColorChooser import askcolor
from supy.guimaker3 import *               # Frame + menu/toolbar builders
from supy.utilities import *
from ivisit.defdb import *
from ivisit.defdialogs import *
from ivisit.widgetdefcollection import *
from ivisit.widgets import *
from ivisit.parser import *
from ivisit.help import *

#from LevelMacher_databases_oldversion import *   # old versions (just for importing data from old versions)
#from LevelMacher_databases import *
#from LevelMacher_displayframe import *
#from SpielMacher_PythonBuilder import *
#PyLevelMacher_widgetdef import *
#from PyLevelMacher_interface import *
#from PyLevelMacher_widgets import *
#from PyLevelMacher_display import *
#from PyLevelMacher_form import *

try:
    import textConfig                      # startup font and colors
    configs = textConfig.__dict__          # work if not on the path or bad
except:
    configs = {}

helptext_ivisit = """IVisit Version %s
--------------------
Interactive VIsual SImulation Tool 
for Python-based generic model simulation and optimization of model parameters

(c) 1/9/2017-25/7/2024, Andreas Knoblauch.
All rights reserved.
EXPERIMENTAL SOFTWARE: 
NO WARRANTIES! REGULARLY BACKUP YOUR DATA!

Author:
Andreas Knoblauch
Hochschule Albstadt-Sigmaringen
Albstadt, Germany

New in version %s:
- default parameter file changed 
  from default.db to <script_name>.db
  thus, (multiple) ivisit scripts can be started reasonably 
  by python <script_name>.py
- new module ivisit.special including SimpleScope 
  for visualization and ClickDragEventAutomaton handling 
  mouse events on GUI widgets
- automatic locks (to ensure mutual exclusion of threads) 
  for calls to init(), bind(), step() by setting flagLock 
  of IVisit_Simulation

Planned for next version: -

Version History:
0.1 : 1/9/2017 GenSim
     - Basic functionality
     - Input/Output Widgets
0.2 : 10/11/2019 
     - Automatic Parsing of simulation parameters 
       and output/display widgets
     - binding of events to widgets
     - reorganized code, bug fixes 
     - renaming from "gensim" (generic simulation framework) 
0.3 : 2/4/2021
     - improved image widgets (with context menu)
     - interface to matplotlib 
     - new widgets (checkbox, radiobutton, button)
     - improved help function
     - bug fixes 
0.4 : 24/7/2022
     - available as PyPI package (install with: pip install ivisit)
     - option to save images from image widgets 
     - modfied text_out widgets (can align text)
     - synchronous update mode for parameter widgets
     - bug fixes 
0.5 : 12/9/2022
     - widget labels can be modified in simulation (setLabelText)
     - support of application help functions
0.51: 1/11/2022
     - extended EPS image export function
     - extended HELP / TITLE functions
0.60: 1/9/2023
     - new parameter-widget class DICTSLIDER + ITEMS
       for numeric parameter-dictionaries
     - bug fixes
0.70: 25/7/2024
     - default parameter file changed 
       from default.db to <script_name>.db
       thus, (multiple) ivisit scripts can be started reasonably 
       by python <script_name>.py
     - new module ivisit.special including SimpleScope 
       for visualization and ClickDragEventAutomaton handling 
       mouse events on GUI widgets
     - automatic locks (to ensure mutual exclusion of threads) 
       for calls to init(), bind(), step() by setting flagLock 
       of IVisit_Simulation
0.71: 30/10/2024
     - bug fixes
"""

START     = '1.0'                          # index of first char: row=1,col=0
SEL_FIRST = SEL + '.first'                 # map sel tag to index
SEL_LAST  = SEL + '.last'                  # same as 'sel.last'
     
FontScale = 0                              # use bigger font on linux
if sys.platform[:3] != 'win':              # and other non-windows boxes
    FontScale = 3


################################################################################
# Main class: implements IVisit gui, actions
################################################################################

class IVisit:                          # mix with menu/toolbar Frame class
    ftypes = [('All files',     '*'),                       # for file open dialog
              ('IVisit database files',   '.db')]           # customize in subclass
     
    colors = [{'fg':'black',      'bg':'white'},      # color pick list
              {'fg':'yellow',     'bg':'black'},      # first item is default
              {'fg':'white',      'bg':'blue'},       # tailor me as desired
              {'fg':'black',      'bg':'beige'},      # or do PickBg/Fg chooser
              {'fg':'yellow',     'bg':'purple'},
              {'fg':'black',      'bg':'brown'},
              {'fg':'lightgreen', 'bg':'darkgreen'},
              {'fg':'darkblue',   'bg':'orange'},
              {'fg':'orange',     'bg':'darkblue'}]
     
    fonts  = [('courier',    9+FontScale, 'normal'),  # platform-neutral fonts
              ('courier',   12+FontScale, 'normal'),  # (family, size, style)
              ('courier',   10+FontScale, 'bold'),    # or popup a listbox
              ('courier',   10+FontScale, 'italic'),  # make bigger on linux
              ('times',     10+FontScale, 'normal'),  # use 'bold italic' for 2
              ('helvetica', 10+FontScale, 'normal'),  # also 'underline', etc.
              ('ariel',     10+FontScale, 'normal'),
              ('system',    10+FontScale, 'normal'),
              ('courier',   20+FontScale, 'normal')]
     
    def __init__(self, loadFirst='',sim=None,debuglevel=0):
        if not isinstance(self, GuiMaker):
            raise TypeError('IVisit needs a GuiMaker mixin')
        if not isinstance(sim, IVisit_Simulation):
            raise TypeError('Parameter sim must be a IVisit_Simulation')
        self.sim=sim            # reference to simulation class
        self.db=None            # database with widget defintions etc.
        self.debuglevel=debuglevel # debug level (default 0, no debug outputs)
        self.wdefcoll=None          # widget definition
        self.display=None       # display frame
        self.steps=0            # simulation step counter
        self.runflg=0           # if 1 then simulation is running automatically
        self.lastfind   = None
        self.openDialog = None
        self.saveDialog = None
        #self.currfile='ivisit_default.db'          # old default file name
        self.currfile=os.path.splitext(os.path.basename(sys.argv[0]))[0]+'.db'  # new default file name: get python script filename and append '.db' 
        #self.text.focus()                          # else must click in text
        print("loadFirst:",loadFirst)
        if loadFirst: 
            self.onOpen(loadFirst)
        else:
            self.loadDatabase(self.currfile)
            loadFirst=self.currfile
        self.setFileName(loadFirst)
        self.updateTimeLabel()
 
    def start(self):                                # run by GuiMaker.__init__
        self.menuBar = [                            # configure menu/toolbar
            ('File', 0,                    # a GuiMaker menu def tree
                 [('Open...',    0, self.onOpen),   # build in method for self
                  ('Save',       0, self.onSave),   # label, shortcut, callback
                  ('Save As...', 5, self.onSaveAs),
                  ('New',        0, self.onNew),
                  'separator',
                  ('Export',     0, self.onExport),
                  'separator',
                  ('Import Databases'       , 0, self.onImportDatabases),
                  'separator',
                  ('Import From Old Version', 0, self.onImportFromOldVersion),
                  'separator',
                  ('Parse From File', 0, self.onParseFromFile),
                  'separator',
                  ('Quit...',    0, self.onQuit)]
            ),
            ('HUBS', 0,                             
                 [('Simulations'             ,0, self.onHUB_Simulation),
                  'separator',
                  ('Edit Parameter Widgets'  ,0, self.onDB_ParameterWidget),
                  ('Edit Data Widgets'       ,0, self.onDB_DataWidget),
                  'separator',
                  ('Edit Parameters'         ,0, self.onDB_Parameter),
                  ('Edit Data Arrays'        ,0, self.onDB_DataArray)]
            ),
            ('Databases', 0,
                 [('Simulations'                             , 0, self.onDB_Simulation          ),
                  ('Parameters'                              , 0, self.onDB_Parameter           ),
                  ('Data Arrays'                             , 0, self.onDB_DataArray           ),
                  'separator',
                  ('Parameter Widgets'                       , 0, self.onDB_ParameterWidget     ),
                  ('Data Widgets'                            , 0, self.onDB_DataWidget          ),
                  ('Comment Widgets'                         , 0, self.onDB_CommentWidget       )]
            ),
            ('Simulation', 0,
                 [('main_init'                               , 0, self.onSim_main_init          ),
                  'separator',
                  ('init'                                    , 0, self.onSim_init               ),
                  ('step'                                    , 0, self.onSim_step               ),
                  'separator',
                  ('run'                                     , 0, self.onSim_run                ),
                  ('stop'                                    , 0, self.onSim_stop               ),
                  ('cont'                                    , 0, self.onSim_cont               )]
            ),
            ('Help', 0,
                 [('Help on IViSiT'                          , 0, self.onHelp_ivisit            ),
                  ('Help on '+str(self.str_app_name)         , 0, self.onHelp_app               ),
                  'separator',
                  ('About IViSiT'                            , 0, self.onAbout_ivisit           ),
                  ('About '+str(self.str_app_name)           , 0, self.onAbout_app              )]
         )]
        self.toolBar = [
            ('Save'             , self.onSave                , {'side': LEFT}),
            ('Init'             , self.onSim_init            , {'side': LEFT}),
            ('Step'             , self.onSim_step            , {'side': LEFT}),
            ('Run'              , self.onSim_run             , {'side': LEFT}),
            ('Stop'             , self.onSim_stop            , {'side': LEFT}),
            ('Cont'             , self.onSim_cont            , {'side': LEFT}),
            ('Simulations'      , self.onHUB_Simulation      , {'side': LEFT}),
            ('Parse'            , self.onParseFromFile       , {'side': LEFT}),
            ('Help'             , self.help                  , {'side': RIGHT}),
            ('Quit'             , self.onQuit                , {'side': RIGHT})]

    def onHelp_ivisit (self): self.help(flagContent='ivisit_help' )
    def onAbout_ivisit(self): self.help(flagContent='ivisit_about')
    def onHelp_app    (self): self.help(flagContent='app_help'    )
    def onAbout_app   (self): self.help(flagContent='app_about'   )
        
    def makeWidgets(self):                          # run by GuiMaker.__init__
        labelframe = Frame(self)
        name = Label(labelframe, bg='black', fg='white')       # add below menu, above tool
        name.pack(side=LEFT, fill=X)                           # menu/toolbars are packed
        timelabel = Label(labelframe, bg='black', fg='white')  # add below menu, above tool
        timelabel.pack(side=RIGHT, fill=X)                     # menu/toolbars are packed
        pointerlabel = Label(labelframe)
        pointerlabel.pack(side=RIGHT,fill=X)
        labelframe.pack(side=TOP, fill=X)

        # main display (for environment)
        #self.display = IVisitDisplay(self,IVisitDisplay_config,
        #                                  db_level_environments,db_being_instances,db_initial_positions,db_movement_paths,
        #                                  db_images, db_being_types, db_beings,
        #                                  pointerlabel=pointerlabel
        #                                  )
        #self.display.pack(side=LEFT,expand=True,fill=BOTH)
        #self.inputmask_db_beings_instances  = None
        #self.inputmask_db_initial_positions = None
        #self.inputmask_db_movement_paths = None
        #self.display.bind('<being-selected>',self.onBeingSelected)
        #self.display.bind('<being-moved>'   ,self.onBeingMoved   )
        #self.display.bind('<initpos-selected>',self.onInitPosSelected)
        #self.display.bind('<intipos-moved>'   ,self.onInitPosMoved   )
        #self.display.bind('<mvpath-selected>',self.onMvPathSelected)
        #self.display.bind('<mvpath-moved>'   ,self.onMvPathMoved   )

        # assign local variables to object fields...
        self.labelframe = labelframe
        self.filelabel = name
        self.timelabel = timelabel
        self.pointerlabel = pointerlabel
        self.display = None


    ############################################################################
    # Project menu commands
    ############################################################################
     
    def my_askopenfilename(self,initialdir=".", initialfile=None):      # objects remember last result dir/file
        if not self.openDialog:
           self.openDialog = Open(initialdir=initialdir,initialfile=initialfile, 
                                  filetypes=self.ftypes)
        return self.openDialog.show()
     
    def my_asksaveasfilename(self,initialdir=".", initialfile=None, title=None):    # objects remember last result dir/file
        if not self.saveDialog:
           self.saveDialog = SaveAs(initialdir=initialdir,initialfile=initialfile, 
                                    filetypes=self.ftypes, title=title)
        return self.saveDialog.show()
        
    def onOpen(self, loadFirst=''):
        print 
        idir,ifile = self.getInitialDirectoryAndFile(loadFirst)
        doit = (not self.db) or (not self.wdefcoll) or (not self.display) or \
               (not self.display.isModified) or askyesno('IVisit', 'Simulation GUI definition has changed: discard changes?')
        if doit:
            filename = loadFirst or self.my_askopenfilename(initialdir=idir, initialfile=ifile)
            if filename:
                self.loadDatabase(filename)
                self.setFileName(filename)
                self.onSim_main_init()

    def loadDatabase(self,filename,debuglevel=None):              # load new simulation database with a given filename
        # load simulation database (or set to default)
        if(debuglevel==None): debuglevel=self.debuglevel
        self.db = sqldatabase(db_ivisit_cfg, filename, debugLevel=debuglevel)    
        simkeys = self.db.simple_select(['key'],['simulation'])
        self.key_sim_display=simkeys[-1][0]            # default simulation most recent simulation (previous: 'EMPTY' simulation)
        self.updateWidgetDefs()
        self.setFileName(self.db.filename)

    def updateWidgetDefs(self,key_sim=None):    # update widget definitions for a simulation with a given key
        if key_sim==None: key_sim=self.key_sim_display
        try:
            self.wdefcoll=IVisitWidgetDefCollection(self.db,self.sim)
            self.wdefcoll.initFromDatabase(key_sim)
        except Exception as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_tb(exc_traceback, limit=None, file=sys.stdout)         
            print("Warning: Exception e=",e)
        self.updateDisplay()

    def updateDisplay(self,key_sim=None):      # update display according to widget definitions 
        if(self.display): 
            self.display.destroy()
        try:
            self.display = IVisitRawDisplayFrame(self,self.wdefcoll)
            self.display.pack(side=LEFT,expand=True,fill=BOTH)
        except Exception as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_tb(exc_traceback, limit=None, file=sys.stdout)         
            print("Warning: Exception e=",e)

    def onSave(self):
        self.onSaveAs(self.currfile)  # may be None
     
    def getInitialDirectoryAndFile(self,forcefile=None):
        if(forcefile==None)or(forcefile==''):
            forcefile=self.getFileName()
        if(forcefile!=None):
            forcefile  = os.path.abspath(forcefile)
            idir,ifile = os.path.split(forcefile)
        else:
            idir,ifile = self.getFileNamePath(), None
        if(idir==None):
            idir=os.path.get_cwd()
        return idir,ifile
     
    def onSaveAs(self, forcefile=None):
        self.display.saveGUIDataToDatabase()             # save GUI changes to old database
        idir,ifile = self.getInitialDirectoryAndFile(forcefile)
        filename = forcefile or self.my_asksaveasfilename(initialdir=idir, initialfile=ifile, title='Save as')
        #if filename != self.currfile:
        #    self.update_filenames_in_databases(self.currfile, filename)
        if filename != self.currfile:  
            if not os.path.exists(filename) or askyesno('IVisit', 'File '+str(filename)+ ' already exists! Do you want to overwrite it?'):
                try:
                    print("copying saved self.currfile=",self.currfile," to filename",filename)
                    shutil.copy2(self.currfile,filename)
                except Exception as e:
                    showerror('IVisit', str(e)+'\nCould not write file ' + filename + '! Keep old file.')
                else:
                    # if all ok do the same procedure as with 'open'
                    self.loadDatabase(filename)
                    self.setFileName(filename)

    def onNew(self, forcefile=None):
        self.display.saveGUIDataToDatabase()             # save GUI changes to old database
        idir,ifile = self.getInitialDirectoryAndFile(forcefile)
        filename = forcefile or self.my_asksaveasfilename(initialdir=idir, initialfile=ifile, title='New')
        if filename:
            if not os.path.exists(filename) or askyesno('IVisit', 'File '+str(filename)+ ' already exists! Do you want to overwrite it?'):
                try:
                    if os.path.exists(filename): os.remove(filename)
                except Exception as e:
                    showerror('IVisit', str(e)+'\nCould not remove file ' + filename + '! Keep file.')
                else:
                    # if all ok do the same procedure as with 'open'
                    self.onOpen(filename)

       
    def onExport(self):
        idir,ifile=self.getInitialDirectoryAndFile()
        ifile=os.path.splitext(ifile)[0]+".py"
        filename=self.my_asksaveasfilename(initialdir=idir, initialfile=ifile)
        if(filename):
            classname       =self.db_level_environments.get_record_entry("Key/Class")
            derivedfromclass=self.db_level_environments.get_record_entry("Derived\nfrom Class")
            build_level_file(filename,classname,derivedfromclass,self.mb_level,None)

    def onImportDatabases(self,list_of_databases = None):
        if list_of_databases == None:
            list_of_databases = ["images.db",
                                 "sounds.db",
                                 "being_types.db",
                                 "being_CHR_collision.db",
                                 "being_CHR_mobility.db",
                                 "being_CHR_character.db",
                                 "being_CHR_question.db",
                                 "being_CHR_ghost.db",
                                 "beings.db"
                                 ]
        idir,ifile=self.getInitialDirectoryAndFile()
        filename=self.my_askopenfilename(initialdir=idir, initialfile=ifile)
        if filename:
            mb1 = Multibase(mb_level_config,filename)
            self.mb_level.import_data_from_multibase(mb1,list_of_databases)

    #def onImportDatabases(self):
    #    idir,ifile=self.getInitialDirectoryAndFile()
    #    filename=self.my_askopenfilename(initialdir=idir, initialfile=ifile)
    #    if filename:
    #        mb1 = Multibase(mb_level_config,filename)
    #        self.importDatabaseFromTo(mb1.get_database("images.db"             ),self.db_images             )
    #        self.importDatabaseFromTo(mb1.get_database("sounds.db"             ),self.db_sounds             )
    #        self.importDatabaseFromTo(mb1.get_database("being_types.db"        ),self.db_being_types        )
    #        self.importDatabaseFromTo(mb1.get_database("being_CHR_collision.db"),self.db_being_CHR_collision)
    #        self.importDatabaseFromTo(mb1.get_database("being_CHR_mobility.db" ),self.db_being_CHR_mobility )
    #        self.importDatabaseFromTo(mb1.get_database("being_CHR_character.db"),self.db_being_CHR_character)
    #        self.importDatabaseFromTo(mb1.get_database("being_CHR_question.db" ),self.db_being_CHR_question )
    #        self.importDatabaseFromTo(mb1.get_database("being_CHR_ghost.db"    ),self.db_being_CHR_ghost    )
    #        self.importDatabaseFromTo(mb1.get_database("beings.db"             ),self.db_beings             )

    #def importDatabaseFromTo(self,db_from, db_to):
    #    keys_from = db_from.get_list_of_keys()
    #    keys_to   = db_to  .get_list_of_keys()
    #    for k in keys_from:
    #        if not (k in keys_to):
    #            r = db_from.get_record(k)
    #            db_to.set_record(*r)

    def onImportFromOldVersion(self,initfile=""):
        if askyesno('Import Level Multibase with old version configuration', 
                    "Choose a filename of an old-version Level file \n"+
                    "according to the multibase configuration given in python file IVisit_databases_oldversion.py.\n"+
                    "Then this old level-file data will be imported to the current level file.\n\n"):
            idir,ifile = self.getInitialDirectoryAndFile(initfile)
            filename_old = self.my_askopenfilename(initialdir=idir, initialfile=ifile)
            if filename_old:
                self.mb_level.cfg.import_leveldata_from_oldversion(self.mb_level,filename_old)
                self.display.synchronize_with_databases()

    def onParseFromFile(self):
        # (i) choose parsing options
        labels           =['Parse file' ,'Simulation name'            ,'Parsing mode'                              ,'Save before parsing?','Select simulation after parsing?']
        parsemode_options=['filename'   ,None                         ,IVisitParser.PARSEMODE_OPTIONS              ,['yes','no']          ,['yes','no']                      ]
        defaults         =[main.__file__,self.wdefcoll.simulation_name,IVisitParser.PARSEMODE_UpdateOldExceptValues,'yes'                 ,'yes'                             ]
        help_event='<Button-3>'
        help_texts=['Choose python simulation file to be parsed\n(default: current python file)',
                    'Name of simulation for which parameters should be parsed\n(default: current simulation; set "" to choose parsed simulation name)',
                    'Choose one of the following parse options:\n'+IVisitParser.parsemode_helptext,
                    'Save all current GUI settings to simulation databases?\n(recommended "yes")',
                    'After parsing, select simulation for which parameters have been parsed?\n(recommended "yes")']
        parseoptions=askSupyForm(labels,defaults,self,"Choose parsing options (press right button for help)...",'e',30,40,options=parsemode_options,help_event=help_event,help_texts=help_texts)
        if parseoptions:
            opt_filename,opt_simname,opt_parsemode,opt_savebefore,opt_selectafter=parseoptions[0],parseoptions[1],parseoptions[2],parseoptions[3],parseoptions[4]
            if opt_savebefore=='yes':     # Save before parsing?
                self.onSave()
            parser = IVisitParser(opt_filename,self.db,opt_parsemode,opt_simname)     
            if opt_selectafter=='yes':    # Select simulation after parsing?
                if opt_simname==self.wdefcoll.simulation_name:   # selected simulation equals current simulation?
                    sim_key=self.key_sim_display                 # then just use current simulation key (prevent using the "wrong" simulation in case of non-unique simulation names!!)
                else:
                    data=self.db.simple_select(['key'],['simulation'],where="name='"+opt_simname+"'")
                    assert len(data)>0, "Simulation name opt_simname="+str(opt_simname)+" cannot be found in database!? Thus, simulation cannot be selected after parsing!!"
                    sim_key=data[0][0]
                self.onHUBSim_SelectSim(None,sim_key)            # select simulation by key

            #idir,ifile = self.getInitialDirectoryAndFile()  # idir is the initial directory
            #ifile = main.__file__                           # name of main python script
            #filename = self.my_askopenfilename(initialdir=idir, initialfile=ifile)
            #if filename:
            #    parser= IVisitParser(filename,self.db) 
        
    def onQuit(self):
        doit = (not self.display is None and not self.display.isModified) \
               or askyesno('IVisit', 'Simulation GUI definition has changed: Quit and discard changes?')    
        if doit:
            self.quit()                 # Frame.quit via GuiMaker

     
    ############################################################################
    # Edit menu commands
    ############################################################################
    def onEditProjectFile(self):
        pass #fillDBInputMask(self,"View and edit database "+db_images_config.fname,db_images,db_images_config)

    def onEditBeingInstances(self):
        self.onDB_BeingInstances()

    def onEditInitialPositions(self):
        self.onDB_InitialPositions()

    def onEditMovementPaths(self):
        self.onDB_MovementPaths()

    def onEditExtraGoals(self):
        self.onDB_ExtraGoals()

    def onEditPlayers(self):
        self.onDB_Players()

    # ----------------

    def onBeingSelected(self,key):
        if self.inputmask_db_beings_instances:
            self.inputmask_db_beings_instances.selectListboxItem(key)

    def onBeingMoved(self,key,pos):
        if self.inputmask_db_beings_instances:
            self.inputmask_db_beings_instances.selectListboxItem(key)

    def onInitPosSelected(self,key):
        if self.inputmask_db_initial_positions:
            self.inputmask_db_initial_positions.selectListboxItem(key)

    def onInitPosMoved(self,key,pos):
        if self.inputmask_db_initial_positions:
            self.inputmask_db_initial_positions.selectListboxItem(key)

    def onMvPathSelected(self,key):
        if self.inputmask_db_movement_paths:
            self.inputmask_db_movement_paths.selectListboxItem(key)

    def onMvPathMoved(self,key,pos):
        if self.inputmask_db_movement_paths:
            self.inputmask_db_movement_paths.selectListboxItem(key)

    ############################################################################
    # Database menu commands
    ############################################################################
    def onDB_images(self):
        fillDBInputMask(self,"View and edit database "+self.db_images.filename,self.db_images,db_images_config,self.mb_level,self.currfile_path)

    def onDB_sounds(self):
        fillDBInputMask(self,"View and edit database "+self.db_sounds.filename,self.db_sounds,db_sounds_config,self.mb_level,self.currfile_path)

    def onDB_CHRCollision(self):
        fillDBInputMask(self,"View and edit database "+self.db_being_CHR_collision.filename,
                        self.db_being_CHR_collision,db_being_CHR_collision_config,self.mb_level,self.currfile_path)

    def onDB_CHRMobility(self):
        fillDBInputMask(self,"View and edit database "+self.db_being_CHR_mobility.filename,
                        self.db_being_CHR_mobility,db_being_CHR_mobility_config,self.mb_level,self.currfile_path)

    def onDB_CHRCharacter(self):
        fillDBInputMask(self,"View and edit database "+self.db_being_CHR_character.filename,
                        self.db_being_CHR_character,db_being_CHR_character_config,self.mb_level,self.currfile_path)

    def onDB_CHRQuestion(self):
        fillDBInputMask(self,"View and edit database "+self.db_being_CHR_question.filename,
                        self.db_being_CHR_question,db_being_CHR_question_config,self.mb_level,self.currfile_path)

    def onDB_CHRGhost(self):
        fillDBInputMask(self,"View and edit database "+self.db_being_CHR_ghost.filename,
                        self.db_being_CHR_ghost,db_being_CHR_ghost_config,self.mb_level,self.currfile_path)

    def onDB_BeingTypes(self):
        fillDBInputMask(self,"View and edit database "+self.db_being_types.filename,self.db_being_types,db_being_types_config,
                        self.mb_level,self.currfile_path)

    def onDB_Beings(self):
        fillBeingsDBInputMask(self,"View and edit database "+self.db_beings.filename,self.db_beings,db_beings_config,
                              self.mb_level,self.currfile_path)

    def onDB_BeingInstances(self):
        top=Toplevel(self)
        top.transient(self)       # make window transient, e.g., minmize with parent etc.
        top.title("View and edit database "+self.db_being_instances.filename) 
        self.inputmask_db_beings_instances=BeingsInstancesDBInputMask(top,self.db_being_instances,db_being_instances_config,
                                                                      self.mb_level,self.currfile_path,
                                                                      self.db_images,self.db_being_types)
        top.wait_window(self.inputmask_db_beings_instances.box)
        self.inputmask_db_beings_instances=None
        top.destroy()

    def onDB_LevelEnvironments(self):
        fillLevelEnvironmentsDBInputMask(self,"View and edit database "+self.db_level_environments.filename,
                                         self.db_level_environments,db_level_environments_config,self.mb_level)

    def onDB_InitialPositions(self):
        top=Toplevel(self)
        top.transient(self)       # make window transient, e.g., minmize with parent etc.
        top.title("View and edit database "+self.db_initial_positions.filename) 
        self.inputmask_db_initial_positions=DBInputMask(top,self.db_initial_positions,db_initial_positions_config,
                                                        self.mb_level,self.currfile_path)
        top.wait_window(self.inputmask_db_initial_positions.box)
        self.inputmask_db_initial_positions=None
        top.destroy()

    def onDB_MovementPaths(self):
        top=Toplevel(self)
        top.transient(self)       # make window transient, e.g., minmize with parent etc.
        top.title("View and edit database "+self.db_movement_paths.filename) 
        self.inputmask_db_movement_paths=DBInputMask(top,self.db_movement_paths,db_movement_paths_config,
                                                     self.mb_level,self.currfile_path)
        top.wait_window(self.inputmask_db_movement_paths.box)
        self.inputmask_db_movement_paths=None
        top.destroy()

    def onDB_ExtraGoals(self):
        fillDBInputMask(self,"View and edit database "+self.db_extra_goals.filename,self.db_extra_goals,db_extra_goals_config,
                        self.mb_level,self.currfile_path)

    def onDB_Players(self):
        fillDBInputMask(self,"View and edit database "+self.db_players.filename,self.db_players,db_players_config,
                        self.mb_level,self.currfile_path)

    ############################################################################
    # Database menu commands
    ############################################################################
    def onDB_Simulation(self):
        editSQLTables(self,self.db,"Editing of Table 'Simulation'",tbed_simulation_cfg)

    def onDB_Parameter(self):
        editSQLTables(self,self.db,"Editing of Table 'Parameter'",tbed_parameter_cfg)

    def onDB_DataArray(self):
        editSQLTables(self,self.db,"Editing of Table 'DataArray'",tbed_dataarray_cfg)

    def onDB_ParameterWidget(self):
        editSQLTables(self,self.db,"Editing of Table 'ParameterWidget'",tbed_parameterwidget_cfg)

    def onDB_DataWidget(self):
        editSQLTables(self,self.db,"Editing of Table 'DataWidget'",tbed_datawidget_cfg)

    def onDB_CommentWidget(self):
        editSQLTables(self,self.db,"Editing of Table 'CommentWidget'",tbed_commentwidget_cfg)

    def onHUB_Simulation(self):
        editSQLTables(self,self.db,"Defining Simulations and their links to Parameters and Data",tbedHUB_simulation_cfg,flagIndepFromParent=1,\
                      extrabuttons=[['Select','listboxleft',self.onHUBSim_SelectSim]])

    def onHUBSim_SelectSim(self,choiceLB,key_sim=None):
        doit = (not self.display.isModified) or askyesno('IVisit', 'Simulation GUI definition has changed: Select simulation and discard changes?')
        if doit:
            #print("choiceLB=",choiceLB)
            #print("selected_data=",choiceLB.selected_data)
            if choiceLB:
                self.key_sim_display=choiceLB.selected_data[0][0]           # simulation key is first entry of data record of listbox
            elif key_sim:
                self.key_sim_display=key_sim
            #print("new key=",self.key_sim_display)
            self.updateWidgetDefs()
            self.updateFileLabel()
            self.updateTimeLabel()

    ############################################################################
    # Simulation menu commands
    ############################################################################
    def onSim_main_init(self):
        if self.sim.flagLock: self.sim.lock.acquire()   # enable thread-safe access to main_init, init(), bind(), step()?
        self.sim.main_init(self) 
        if self.sim.flagLock: self.sim.lock.release()   # release lock again
        self.onSim_init()
        #self.steps=0
        #self.updateTimeLabel()

    def onSim_init(self):
        if not self.display is None: self.display.resetButtonsAndFlags(1)   # mark parameter widgets as modified (to initialize all relevant data fields in step())
        if self.sim.updateMode=='sync': self.wdefcoll.setAllSimulationParameters()  # set all simulation parameters from widgets
        if self.sim.flagLock: self.sim.lock.acquire()   # enable thread-safe access to main_init, init(), bind(), step()?
        self.sim.init()
        self.sim.bind(self,self.display)
        if self.sim.flagLock: self.sim.lock.release()   # release lock again
        self.steps=0
        self.updateTimeLabel()

    def onSim_step(self):
        if(self.runflg==0):
            self.doOneStepCycle()
        else:
            self.runflg=0

    def onSim_run(self):
        if(self.runflg==0):
            self.onSim_init()
            self.onSim_cont()
        else:
            self.runflg=0

    def onSim_stop(self):                           # stop simulation
        if(self.runflg==1):
            self.runflg=0

    def doOneStepCycle(self):
        for i in range(self.wdefcoll.simulation_frames_per_step):          # loop over number per display frames per step call
            for j in range(self.wdefcoll.simulation_simsteps_per_frame):
                if self.sim.updateMode=='sync': self.wdefcoll.setAllSimulationParameters()
                self.steps+=1
                if self.sim.flagLock: self.sim.lock.acquire()   # enable thread-safe access to main_init, init(), bind(), step()?
                self.sim.step()
                if self.sim.flagLock: self.sim.lock.release()   # release lock again
                self.display.resetButtonsAndFlags()   # reset Button variables after each call to step() (where the values may be checked!!)
            #self.steps=self.steps+self.wdefcoll.simulation_simsteps_per_frame
            self.wdefcoll.setAllSimulationDataArrays()
            self.display.updateData()   #updateDisplay()
            self.updateTimeLabel()
            #self.steps=self.libSim.PyFeInter_step(self.widgetdefs.felixsteps_per_frame)
            #self.display.updateData()
            #self.setSimTime()
            #if i<(self.widgetdefs.frames_per_step-1):self.update_idletasks()    # force display of each step ...
        
    def onSim_cont(self):                           # continuous simulation (start and stop)
        self.runflg=1-self.runflg;
        self.step_repeater()

    def step_repeater(self):
        if(self.runflg==1):
            self.doOneStepCycle()
            #print("repeater: delay=",self.wdefcoll.simulation_delay_per_step)
            self.after(self.wdefcoll.simulation_delay_per_step, self.step_repeater)
            


    ############################################################################
    # Simulation menu commands
    ############################################################################

    def onFigure(self):                         # draw data etc.
        if(self.data!=None):
            w = Toplevel(self)
            t = matplotlib.dates.date2num([datetime.datetime.strptime(s,"%m/%d/%Y") for s in self.data['date']])
            v = self.data['close']
            cfg = IVisitPlotFrame_config()
            cfg.x=t
            cfg.y=v
            t_disp_min = matplotlib.dates.num2date(t[-1]).date()-datetime.timedelta(days=90)
            t_disp_max = matplotlib.dates.num2date(t[-1]).date()+datetime.timedelta(days=90)
            t_min      = matplotlib.dates.num2date(t[-1]).date()-datetime.timedelta(days=60)
            t_max      = matplotlib.dates.num2date(t[-1]).date()+datetime.timedelta(days=0)
            t_pred_min = matplotlib.dates.num2date(t[-1]).date()-datetime.timedelta(days=60)
            t_pred_max = matplotlib.dates.num2date(t[-1]).date()+datetime.timedelta(days=30)
            cfg.t_disp_min, cfg.t_disp_max = matplotlib.dates.date2num(t_disp_min), matplotlib.dates.date2num(t_disp_max)
            cfg.t_min     , cfg.t_max      = matplotlib.dates.date2num(t_min)     , matplotlib.dates.date2num(t_max)
            cfg.t_pred_min, cfg.t_pred_max = matplotlib.dates.date2num(t_pred_min), matplotlib.dates.date2num(t_pred_max)
            cfg.mv_avg_window=1
            cfg.sampling_interval=1
            cfg.regr_order, regr_type = 1, 'ml'
            cfg.trafo_flag, cfg.trafo_t1, cfg.trafo_t2 = 1, -5,5
            cfg.extremum_domination=10
            f = IVisitPlotFrame(w,cfg).pack(fill=BOTH, expand=1)
            
    def updateFigure(self):
        t = matplotlib.dates.date2num([datetime.datetime.strptime(s,"%m/%d/%Y") for s in self.data['date']])
        v = self.data['close']
        plt.plot_date(t,v,'k-')
        plt.grid()
        plt.show()


    def onRun(self):                            # run simulation
        if(self.runflg==0):
            self.onInit()
            self.onCont()
        else:
            self.runflg=0

    def onStop(self):                           # stop simulation
        if(self.runflg==1):
            self.runflg=0

    def onInit(self):                           # initialize simulation
        self.libSim.PyFeInter_init()
        self.steps=0
        self.setSimTime()

    def onStep(self):                           # do one simulation step
        if(self.runflg==0):
            self.doOneStepCycle()
        else:
            self.runflg=0

    #def doOneStepCycle(self):
    #    for i in range(self.wdefcoll.frames_per_step):              # loop over number per display frames per step call
    #        self.steps=self.libSim.PyFeInter_step(self.widgetdefs.felixsteps_per_frame)
    #        self.display.updateData()
    #        self.setSimTime()
    #        if i<(self.widgetdefs.frames_per_step-1):self.update_idletasks()    # force display of each step ...
        
    #def onCont(self):                           # continuous simulation (start and stop)
    #    self.runflg=1-self.runflg;
    #    self.step_repeater()

    #def step_repeater(self):
    #    if(self.runflg==1):
    #        self.doOneStepCycle()
    #        self.after(self.widgetdefs.delay_per_step, self.step_repeater)
            
    def onControl(self):                           # edit control parameters
        labels=['simulation steps per frame','frames per cycle','delay per cycle [msec]']
        defaults=[str(self.widgetdefs.felixsteps_per_frame), \
                  str(self.widgetdefs.frames_per_step),\
                  str(self.widgetdefs.delay_per_step)]
        r=askPyFeForm(labels,defaults,self,"Edit simulation control parameters",'w',labelwidth=None,entrysize=20)
        if r:
            self.widgetdefs.felixsteps_per_frame = int(r[0])
            self.widgetdefs.frames_per_step      = int(r[1])
            self.widgetdefs.delay_per_step       = int(r[2])

    def onReparse(self):                           # reparse Felix simulation parameters
        self.runflg=1-self.runflg;
        self.step_repeater()

    def onUndo(self):                           # 2.0
        try:                                    # tk8.4 keeps undo/redo stacks
            pass #self.text.edit_undo()               # exception if stacks empty
        except TclError:                        # menu tear-offs for quick undo
            showinfo('PyEdit', 'Nothing to undo')
            
    def onRedo(self):                           # 2.0: redo an undone
        try:
            pass #self.text.edit_redo()
        except TclError:
            showinfo('PyEdit', 'Nothing to redo')
        
    def onCopy(self):                           # get text selected by mouse,etc
        #if not self.text.tag_ranges(SEL):       # save in cross-app clipboard
        #    showerror('PyEdit', 'No text selected')
        #else:
        #    text = self.text.get(SEL_FIRST, SEL_LAST)  
        #    self.clipboard_clear()              
        #    self.clipboard_append(text)
        pass
    
    def onDelete(self):                         # delete selected text, no save
        pass
        #if not self.text.tag_ranges(SEL):
        #    showerror('PyEdit', 'No text selected')
        #else:
        #    self.text.delete(SEL_FIRST, SEL_LAST)
     
    def onCut(self):
        pass
        #if not self.text.tag_ranges(SEL):
        #    showerror('PyEdit', 'No text selected')
        #else: 
        #    self.onCopy()                       # save and delete selected text
        #    self.onDelete()
     
    def onPaste(self):
        pass
        #try:
        #    text = self.selection_get(selection='CLIPBOARD')
        #except TclError:
        #    showerror('PyEdit', 'Nothing to paste')
        #    return
        #self.text.insert(INSERT, text)          # add at current insert cursor
        #self.text.tag_remove(SEL, '1.0', END) 
        #self.text.tag_add(SEL, INSERT+'-%dc' % len(text), INSERT)
        #self.text.see(INSERT)                   # select it, so it can be cut
     
    def onSelectAll(self):
        pass
        #self.text.tag_add(SEL, '1.0', END+'-1c')   # select entire text 
        #self.text.mark_set(INSERT, '1.0')          # move insert point to top
        #self.text.see(INSERT)                      # scroll to top

    ############################################################################
    # Search menu commands
    ############################################################################
 
    def onGoto(self, forceline=None):
        pass
        #line = forceline or askinteger('PyEdit', 'Enter line number')
        #self.text.update() 
        #self.text.focus()
        #if line is not None:
        #    maxindex = self.text.index(END+'-1c')
        #    maxline  = int(maxindex.split('.')[0])
        #    if line > 0 and line <= maxline:
        #        self.text.mark_set(INSERT, '%d.0' % line)      # goto line
        #        self.text.tag_remove(SEL, '1.0', END)          # delete selects
        #        self.text.tag_add(SEL, INSERT, 'insert + 1l')  # select line
        #        self.text.see(INSERT)                          # scroll to line
        #    else:
        #        showerror('PyEdit', 'Bad line number')
     
    def onFind(self, lastkey=None):
        pass
        #key = lastkey or askstring('PyEdit', 'Enter search string')
        #self.text.update()
        #self.text.focus()
        #self.lastfind = key
        #if key:                                                    # 2.0: nocase
        #    nocase = configs.get('caseinsens', 1)                  # 2.0: config
        #    where = self.text.search(key, INSERT, END, nocase=nocase)
        #    if not where:                                          # don't wrap
        #        showerror('PyEdit', 'String not found')
        #    else:
        #        pastkey = where + '+%dc' % len(key)           # index past key
        #        self.text.tag_remove(SEL, '1.0', END)         # remove any sel
        #        self.text.tag_add(SEL, where, pastkey)        # select key 
        #        self.text.mark_set(INSERT, pastkey)           # for next find
        #        self.text.see(where)                          # scroll display
     
    def onRefind(self):
        pass
        #self.onFind(self.lastfind)
     
    def onChange(self):
        pass
        #new = Toplevel(self)
        #Label(new, text='Find text:').grid(row=0, column=0)
        #Label(new, text='Change to:').grid(row=1, column=0)
        #self.change1 = Entry(new)
        #self.change2 = Entry(new)
        #self.change1.grid(row=0, column=1, sticky=EW)
        #self.change2.grid(row=1, column=1, sticky=EW)
        #Button(new, text='Find',  
        #       command=self.onDoFind).grid(row=0, column=2, sticky=EW)
        #Button(new, text='Apply', 
        #       command=self.onDoChange).grid(row=1, column=2, sticky=EW)
        #new.columnconfigure(1, weight=1)    # expandable entrys
     
    def onDoFind(self):
        pass
        #self.onFind(self.change1.get())                    # Find in change box
     
    def onDoChange(self):
        pass
        #if self.text.tag_ranges(SEL):                      # must find first
        #    self.text.delete(SEL_FIRST, SEL_LAST)          # Apply in change
        #    self.text.insert(INSERT, self.change2.get())   # deletes if empty
        #    self.text.see(INSERT)
        #    self.onFind(self.change1.get())                # goto next appear
        #    self.text.update()                             # force refresh
     
    ############################################################################
    # Tools menu commands 
    ############################################################################
     
    def onFontList(self):
        pass
        #self.fonts.append(self.fonts[0])           # pick next font in list
        #del self.fonts[0]                          # resizes the text area
        #self.text.config(font=self.fonts[0]) 
     
    def onColorList(self):
        pass
        #self.colors.append(self.colors[0])         # pick next color in list
        #del self.colors[0]                         # move current to end
        #self.text.config(fg=self.colors[0]['fg'], bg=self.colors[0]['bg']) 
     
    def onPickFg(self):
        pass
        #self.pickColor('fg')                       # added on 10/02/00
    def onPickBg(self):                             # select arbitrary color
        pass
        #self.pickColor('bg')                       # in standard color dialog

    def pickColor(self, part):                      # this is too easy
        pass
        #(triple, hexstr) = askcolor()
        #if hexstr:
        #    self.text.config(**{part: hexstr})
     
    def onInfo(self):
        pass
        #text  = self.getAllText()                  # added on 5/3/00 in 15 mins
        #bytes = len(text)                          # words uses a simple guess: 
        #lines = len(text.split('\n'))              # any separated by whitespace
        #words = len(text.split()) 
        #index = self.text.index(INSERT)
        #where = tuple(index.split('.'))
        #showinfo('PyEdit Information',
        #         'Current location:\n\n' +
        #         'line:\t%s\ncolumn:\t%s\n\n' % where +
        #         'File text statistics:\n\n' +
        #         'bytes:\t%d\nlines:\t%d\nwords:\t%d\n' % (bytes, lines, words))
     
    def onClone(self):
        new = Toplevel()                # a new edit window in same process
        myclass = self.__class__        # instance's (lowest) class object
        myclass(new)                    # attach/run instance of my class
     
    def onRunCode(self, parallelmode=True):
        """
        run Python code being edited--not an ide, but handy;
        tries to run in file's dir, not cwd (may be PP3E root);
        inputs and adds command-line arguments for script files;
        code's stdin/out/err = editor's start window, if any:
        run with a console window to see code's print outputs;
        but parallelmode uses start to open a dos box for i/o;
        module search path will include '.' dir where started;
        in non-file mode, code's Tk root window is PyEdit win;
        """
        def askcmdargs():
            return askstring('PyEdit', 'Commandline arguments?') or ''
        
        from PP3E.launchmodes import System, Start, Fork
        filemode = False
        thefile  = str(self.getFileName())
        if os.path.exists(thefile):
            filemode = askyesno('PyEdit', 'Run from file?')
        if not filemode:                                    # run text string
            cmdargs   = askcmdargs()
            namespace = {'__name__': '__main__'}            # run as top-level
            sys.argv  = [thefile] + cmdargs.split()         # could use threads
            exec(self.getAllText() + '\n' in namespace)     # exceptions ignored
        elif self.spielmacher_edit_modified:                     # 2.0: changed test
            showerror('PyEdit', 'Text changed: save before run')
        else:
            cmdargs  = askcmdargs()
            mycwd    = os.getcwd()                          # cwd may be root
            os.chdir(os.path.dirname(thefile) or mycwd)     # cd for filenames
            thecmd   = thefile + ' ' + cmdargs
            if not parallelmode:                            # run as file
                System(thecmd, thecmd)()                    # block editor
            else:
                if sys.platform[:3] == 'win':               # spawn in parallel
                    Start(thecmd, thecmd)()                 # or use os.spawnv
                else:
                    Fork(thecmd, thecmd)()                  # spawn in parallel
            os.chdir(mycwd)

    def onPickFont(self):
        pass
        # 2.0 font spec dialog
        #new = Toplevel(self)
        #Label(new, text='Family:').grid(row=0, column=0)      # nonmodal dialog
        #Label(new, text='Size:  ').grid(row=1, column=0)      # see pick list 
        #Label(new, text='Style: ').grid(row=2, column=0)      # for valid inputs
        #self.font1 = Entry(new)
        #self.font2 = Entry(new)
        #self.font3 = Entry(new)
        #self.font1.insert(0, 'courier')                       # suggested vals
        #self.font2.insert(0, '12')
        #self.font3.insert(0, 'bold italic')
        #self.font1.grid(row=0, column=1, sticky=EW)
        #self.font2.grid(row=1, column=1, sticky=EW)
        #self.font3.grid(row=2, column=1, sticky=EW)
        #Button(new, text='Apply', 
        #       command=self.onDoFont).grid(row=3, columnspan=2)
        #new.columnconfigure(1, weight=1)    # expandable entrys

    def onDoFont(self):
        pass
        #try:
        #    font = (self.font1.get(), int(self.font2.get()), self.font3.get())
        #    self.text.config(font=font) 
        #except:
        #    showerror('PyEdit', 'Bad font specification')
     
    ############################################################################
    # Utilities, useful outside this class
    ############################################################################
     
    #def isEmpty(self):
    #    return not self.getAllText() 
     
    #def getAllText(self):
    #    return self.text.get('1.0', END+'-1c')  # extract text as a string     
    #def setAllText(self, text):
    #    self.text.delete('1.0', END)            # store text string in widget
    #    self.text.insert(END, text)             # or '1.0'
    #    self.text.mark_set(INSERT, '1.0')       # move insert point to top 
    #    self.text.see(INSERT)                   # scroll to top, insert set
    #def clearAllText(self):
    #    self.text.delete('1.0', END)            # clear text in widget 
     
    def getFileName(self):                       # full file name
        return self.currfile
    def getFileNameFile(self):                   # only file name without path
        return self.currfile_file
    def getFileNamePath(self):                   # only path
        return self.currfile_path
    def setFileName(self, name):                 # also: onGoto(linenum)
        self.currfile = name  # for save
        if(self.currfile!=None)and(self.currfile!=""):
            self.currfile = os.path.abspath(self.currfile)
            self.currfile_path, self.currfile_file = os.path.split(self.currfile)
        else:
            self.currfile_path, self.currfile_file = "", None
        self.updateFileLabel()
       
    def updateFileLabel(self):
        if(self.currfile==None) or (self.currfile==""):
            self.filelabel.configure(text='None')
        else:
            self.filelabel.configure(text=self.currfile+"::"+str(self.key_sim_display)+"::"+self.wdefcoll.simulation_name)
        
    def updateTimeLabel(self):
        if(self.steps==None) or (self.steps==""):
            self.timelabel.configure(text='None')
        else:
            self.timelabel.configure(text="steps="+str(self.steps))
        

    #def setBg(self, color):
    #    self.text.config(bg=color)              # to set manually from code
    #def setFg(self, color):
    #    self.text.config(fg=color)              # 'black', hexstring
    #def setFont(self, font):
    #    self.text.config(font=font)             # ('family', size, 'style')
        
    #def setHeight(self, lines):                 # default = 24h x 80w
    #    self.text.config(height=lines)          # may also be from textCongif.py
    #def setWidth(self, chars):
    #    self.text.config(width=chars)

    #def clearModified(self):
    #    self.text.edit_modified(0)              # clear modified flag
    def isModified(self):
        return self.spielmacher_edit_modified    # changed since last reset?
 
    def help(self,flagContent='ivisit_about'):
        htext,htitle = "No Help Text Available!","Help"
        if flagContent=='ivisit_help':
            htext=readme_text_ivisit
            htitle='Help IVisit'
        elif flagContent=='ivisit_about':
            htext=helptext_ivisit % ((Version,)*2)+"\n\n"+readme_text_ivisit
            htitle='About IVisit'
        elif flagContent=='app_help':
            htext=str(self.str_app_help)
            htitle='Help '+str(self.str_app_name) 
        elif flagContent=='app_about':
            htext=str(self.str_app_about)
            htitle='About '+str(self.str_app_name) 
        #showinfo('About IVisit', helptext % ((Version,)*2))
        askSupyTextForm(default_text=htext,parent=self,title=htitle,readonly=True,height=40,width=80, wrap=None)
     
################################################################################
# ready-to-use editor classes 
# mix in a Frame subclass that builds menu/toolbars
################################################################################

#    
# when IVisit owns the window 
#
class IVisitMain(IVisit, GuiMakerWindowMenu):  # add menu/toolbar maker 
    def __init__(self, parent=None, loadFirst='', sim=None, debuglevel=0, str_app_name=None, str_app_help=None, str_app_about=None):     # when fills whole window
        self.str_app_name=str_app_name
        self.str_app_help=str_app_help
        self.str_app_about=str_app_about
        self.helpButton=0                               # avoid extra help menu 
        GuiMaker.__init__(self, parent) #,str_appname,str_apphelp)                 # use main window menus
        IVisit.__init__(self, loadFirst,sim,debuglevel) # self has GuiMaker frame
        str_title='IVisit ' + Version
        if not str_app_name is None: str_title=str_app_name
        self.master.title(str_title)                    # title if stand-alone
        self.master.iconname('IVisit')                  # catch wm delete button
        self.master.protocol('WM_DELETE_WINDOW', self.onQuit)
     
class IVisitMainPopup(IVisit, GuiMakerWindowMenu):
    def __init__(self, parent=None, loadFirst='', sim=None, winTitle='', debuglevel=0):     
        self.popup = Toplevel(parent)                  # create own window
        GuiMaker.__init__(self, self.popup)            # use main window menus
        IVisit.__init__(self, loadFirst,sim,debuglevel) 
        assert self.master == self.popup
        self.popup.title('IVisit ' + Version + winTitle) 
        self.popup.iconname('IVisit')               
    def quit(self):
        self.popup.destroy()                           # kill this window only

#         
# when embedded in another window
#
class IVisitComponent(IVisit, GuiMakerFrameMenu):     
    def __init__(self, parent=None, loadFirst='', sim=None, debuglevel=0):     # use Frame-based menus
        GuiMaker.__init__(self, parent)                # all menus, buttons on
        IVisit.__init__(self, loadFirst,sim,debuglevel)           # GuiMaker must init 1st
     
class IVisitComponentMinimal(IVisit, GuiMakerFrameMenu): 
    def __init__(self, parent=None, loadFirst='', sim=None, deleteFile=1, debuglevel=0):   
        self.deleteFile = deleteFile
        GuiMaker.__init__(self, parent)             
        IVisit.__init__(self, loadFirst,sim,debuglevel) 
    def start(self):
        IVisit.start(self)                         # GuiMaker start call
        for i in range(len(self.toolBar)):             # delete quit in toolbar
            if self.toolBar[i][0] == 'Quit':           # delete file menu items
                del self.toolBar[i]; break             # or just disable file
        if self.deleteFile:
            for i in range(len(self.menuBar)):
                if self.menuBar[i][0] == 'File':
                    del self.menuBar[i]; break
        else:
            for (name, key, items) in self.menuBar:
                if name == 'File':
                    items.append([1,2,3,4,6]) 

################################################################################
# stand-alone program run
################################################################################
                                                     
def testPopup():     
    # see PyView and PyMail for component tests
    root = Tk()
    IVisitMainPopup(root)
    IVisitMainPopup(root)
    Button(root, text='More', command=IVisitMainPopup).pack(fill=X)
    Button(root, text='Quit', command=root.quit).pack(fill=X)
    root.mainloop()
     
def IVisit_main(fname=None,sim=None,debuglevel=0,str_app_name=None,str_app_help=None,str_app_about=None):  # may be typed or clicked
    try:                                              # or associated on Windows
        if(fname==None): fname = sys.argv[1]          # arg = optional filename
    except IndexError:
        fname = None
    print("starting with fname=",fname)
    if sim==None: sim=IVisit_Simulation()
    ivm=IVisitMain(loadFirst=fname,sim=sim,debuglevel=debuglevel,str_app_name=str_app_name,str_app_help=str_app_help,str_app_about=str_app_about)
    ivm.pack(expand=YES, fill=BOTH)
    sim.main_init(parent=ivm)
    mainloop()
     
if __name__ == '__main__':                            # when run as a script
    #testPopup()
    IVisit_main()                                     # run .pyw for no dos box    



#from Tkinter import *
import threading     # just for realizing locks to protect processing in init(),step() etc. from mouse event processing
import numpy as np
from copy import *

# *************************************************************************
# Interface IVisit_ParameterSet and IVisit_Simulation:
# Provides Structure for a simulation that can be controlled by IVisit 
# *************************************************************************
class IVisit_Parameters:        # Example Struct of Simulation parameters (are referred to via database objects Parameter.name)
    par1 = 1                           # int parameter
    par2 = 2.5                         # float parameter
    par3 = 'flag_on'                   # text parameter

class IVisit_Data:              # Example Struct of Data Fields to be displayed (that are referred to via database objects DataArray.name)
    data1 = np.array([1])              # single int data
    data2 = np.array([2.2])            # single float data
    data3 = ['textdata']               # single text data
    data4 = np.array([[1,2],[3,4]])    # numerical data matrix

class IVisit_Simulation:        # derive your simulation from this class (do not forget to call the constructor!!!!)
    def __init__(self,name_arg="IVisitSimulation1", parameters_arg=IVisit_Parameters, data_arg=IVisit_Data, updateMode='sync', flagLock=False):
        """
        :param updateMode: if updateMode=='sync' then parameter widgets (e.g., sliders) are written to simulation only before call to step()
                           othersie (updateMode=='async') parameter widgets are written through to simulation immediately after changes
        """
        # store name, parameters, data fields
        self.name=name_arg
        self.parameters=parameters_arg
        self.data=data_arg
        self.updateMode=updateMode
        # create a parameter dictionaries of the simulation parameter and simulation data
        self.parameters_dict          = self.get_dict_of_parameters()           # get working copy of the parameters
        self.parameters_dict_original = self.get_dict_of_parameters(1) # get a deep copy to retain original parameter values (during initializiation)
        # create a dictionary of the simulation data attributes
        self.data_dict = self.get_dict_of_data()             # get copy of data (no deep copy) containing references to the simulation data
        self.display=None                                    # reference to display object (if needed)
        self.parent=None                                     # reference to parent object (if needed)
        # create lock to ensure thread-safe mouse events
        self.lock=threading.Lock()    # lock to protect data processing in init(), bind(), and step() from processing mouse events in separate threads 
        self.flagLock=flagLock        # if True then use lock for init(), bind(), step() etc. 
        
    def get_dict_of_parameters(self,flagDeepCopy=0,parameters=None):    # returns a dict of the (self.)parameters
        if parameters==None: parameters=self.parameters
        pars_dict=vars(parameters)                                                                             # basic dict of attributes of the parameter struct
        pars_dict={k:pars_dict[k] for k in pars_dict.keys() if not(k.startswith('__') and k.endswith('__'))}   # refine (to keep only the real simulation parameters) 
        if flagDeepCopy>0:
            pars_dict = {k:deepcopy(pars_dict[k]) for k in pars_dict.keys()}                                   # make a deep copy of all attributes (to preserve lists etc.)
        return pars_dict

    def get_dict_of_data(self,flagDeepCopy=0,data=None):                # returns a dict of the (self.)data
        if data==None: data=self.data
        data_dict=vars(data)                                                                                   # basic dict of attributes of the data struct
        data_dict={k:data_dict[k] for k in data_dict.keys() if not(k.startswith('__') and k.endswith('__'))}   # refine (to keep only the real simulation data) 
        if flagDeepCopy>0:
            data_dict = {k:deepcopy(data_dict[k]) for k in data_dict.keys()}                                   # make a deep copy of all attributes (to preserve lists etc.)
        return data_dict

    def set_dict_of_parameters(self,pars_dict=None):      # set parameters with the values of pars_dict 
        if pars_dict==None: pars_dict=self.parameters_dict
        self.parameters_dict=pars_dict
        for k in pars_dict:
            setattr(self.parameters,k,pars_dict[k])

    def print_parameters(self,pars=None,indent=""):
        if(pars==None): pars=self.parameters
        allpars=vars(pars)         
        allpars={k:allpars[k] for k in allpars.keys() if not(k.startswith('__') and k.endswith('__'))}
        for k in allpars.keys():
            print(str(indent)+"Parameter '"+str(k)+"': "+str(allpars[k]))

    def isModified(self,widgetname):
        if self.display is None: return 0
        return self.display.getFlagModified(widgetname)
            
    def print_data(self,data=None,indent=""):
        if(data==None): data=self.data
        alldata=vars(data)         
        alldata={k:alldata[k] for k in alldata.keys() if not(k.startswith('__') and k.endswith('__'))}
        for k in alldata.keys():
            print(str(indent)+"Data field '"+str(k)+"': "+str(alldata[k]))

    def print_state(self):
        print("IVisit_Simulation "+str(self.name)+":")
        print("   parameters = "+str(self.parameters))
        print("   data       = "+str(self.data))
        print("   parameters_dict          = "+str(self.parameters_dict))
        print("   parameters_dict_original = "+str(self.parameters_dict_original))
        print("   data_dict  = "+str(self.data_dict))
        print("   Parameter Struct = ...")
        self.print_parameters(indent="      ")
        print("   Data Struct = ...")
        self.print_data(indent="      ")

    def main_init(self,parent=None): # main init procedure (to initialize/booting by start...)
        self.parent=parent

    def init(self):           # init a new simulation run (but keep allocated data)
        pass

    def bind(self,parent=None,display=None):  # bind events to the simulation; parent is typcially the ivisit object (having a display object where to bind to)
        pass

    def step(self):           # one simulation step
        pass

# ****************************************************************************************************************************

if __name__ == '__main__':
    print("\nModule test of ivisit.simulation.py")
    print("--------------------------------------\n") 

    class MyParameters:
        par1 = 1                     # int parameter
        par2 = 2.5                   # float parameter
        par3 = 'flag_on'             # text parameter
        size = 100;                  # size of matrix
        listpar = [6,8,9]            # list parameter
        listpar2 = [1,2,3]           # another list parameter
        
    class MyData:              # Example Struct of Data Fields to be displayed (that are referred to via database objects DataArray.name)
        data1 = np.array([1])              # single int data
        data2 = np.array([2.2])            # single float data
        data3 = ['textdata']               # single text data
        data4 = np.array([[1,2],[3,4]])    # numerical data matrix
        mat = np.array([[1,2],[3,4]])      # another data matrix 

    class MySimulation(IVisit_Simulation):
        def __init__(self,name_arg="MySimulation1",parameters_arg=MyParameters,data_arg=MyData):
            IVisit_Simulation.__init__(self,name_arg,parameters_arg,data_arg)

        def main_init(self): pass          # or further stuff...

        def init(self): pass               # or further stuff...

        def step(self): 
            p=self.parameters
            self.mat = np.random.rand(p.size,p.size)  # generate random matrix

    sim1 = MySimulation("MySimulation1",MyParameters,MyData)
    sim1.main_init()
    sim1.init()
    sim1.step()
    sim1.print_state()
    
    # do a parameter change (as may be done by ivisit)
    pars=sim1.get_dict_of_parameters()
    pars['par1']=1000
    pars['listpar']=[100,101,102]
    pars['listpar2'][0]=10000
    sim1.set_dict_of_parameters(pars)

    print("\nState after parameter changes:")
    sim1.print_state()




#!/usr/bin/python
# -*- coding: utf-8-unix -*-

from supy.sqlforms import *
from supy.sqltableeditor import *
from ivisit.defdb import *

##################################################################################################################################################################
##################################################################################################################################################################
##################################################################################################################################################################
#        
# Define Dialogs for IVISIT (to edit databases defined in ivisit.defdb)      
#        
##################################################################################################################################################################
##################################################################################################################################################################
##################################################################################################################################################################


##################################################################################################################################################################
##################################################################################################################################################################
#        
# Part I: Base class configuration of dialog widgets 
#        
##################################################################################################################################################################
##################################################################################################################################################################

##################################################################################################################################################################
# I.1: Single-Choice Listbox 
##################################################################################################################################################################
class lb_singlechoice_cfg(SQLListboxForm_config):
    # class fields from SQLListbox_config
    tables = None
    join_on = None
    cols   = None
    cols_format = None
    cols_sortbutton = None
    cols_sortbutton = None
    where = None
    where_init = None
    labeltext = None
    colvals2str = colvals2str_default
    sep = ' | '
    width = 45
    height = 10
    lbfont = ('courier',12,'normal')
    callback_select = callback_select_default
    mode_multiselect = False
    # class fields from SQLListboxForm_config
    button_str = ['Ok','Abbrechen']
    title      = 'SQLListboxForm...'  # is used only for askSQLListboxForm

##################################################################################################################################################################
# I.2: Multi-Choice Listbox 
##################################################################################################################################################################
class lb_multichoice_cfg(lb_singlechoice_cfg): 
    mode_multiselect = True

##################################################################################################################################################################
# I.3: SQL Input Form SupySQLForm_config
##################################################################################################################################################################
class form_cfg(SupySQLForm_config):
    tables                = None
    join_on               = None
    cols                  = None
    cols_type             = None
    cols_ref              = None
    cols_readonly         = None
    cols_label            = None
    cols_label_pos        = None
    cols_label_width      = 20     # may also be list
    cols_label_anchor     = 'w'    # may also be list
    cols_pos              = None
    cols_size             = 40     # may also be list
    cols_helptext         = None
    cols_browsebutton_pos = None
    select_cols           = None
    select_vals           = None
    help_event            = "<Button-3>"
    help_title            = "Hilfe"
    help_buttonname       = "Schließen"
    browse_buttonname     = "Auswählen"
    checkResults          = SupySQLForm_checkResults_default  # function to check if inputs are valid (see SupySQLForm_checkResults_default)
    update_tables         = None                              # tables to be updated after form submission
    update_tables_pkeys   = None                              # for each table (to be updated) a list of primary keys 
    update_tables_cols    = None                              # for each table (to be updated) a list of columns to be updated
 
##################################################################################################################################################################
# I.4: SQL Table Editor Input Form SupySQLForm_config
##################################################################################################################################################################
class tbed_config(SQLTableEditor_config):
    cfg_choiceLB         = None               # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm          = None               # configuration for record form (typically a SupySQLForm_config)
    align_widgets        = 'horizontal'       # either 'vertical' or 'horizontal' for alignment of listbox and form
    lbButtons_str        = ['Neu', 'Kopieren', 'Löschen']            # text for listbox buttons
    recFormButtons_str   = ['Zurücksetzen', 'Standardwerte', 'Bestätigen']       # text for form buttons
    ctrlRadioButtons_str = ['Nur anschauen', 'Verändern']                     # text for radiobuttons (read-only, write/edit option)
    ctrlButtons_str      = ['Bestätigen & Schließen', 'Abbrechen & Schließen'] # text for control buttons
    text_showinfo_cannot_delete_readonly = ("Information", "Datensatz kann nicht gelöscht werden! Nur Lese-Zugriff erlaubt!")
    text_askyesno_delete = ("Datensatz löschen?", "Wollen Sie den folgenden Datensatz wirklich löschen? \nDatensatz : ")
    # extend tbed_.... to get a valid initialization of a new record that is conditioned on exam.key (because examofcourse has no direct link to exam)
    ntom_init_on_new     = None    # list of [col_to_be_initialized,col_initialized_by,where_cond]
    

##################################################################################################################################################################
# I.5: RefNtoM Simple Listbox/Textframe 
##################################################################################################################################################################
class ntm_simple(SupySQLntomTextFrame_cfg,lb_singlechoice_cfg):
    readonly=True
    title      = None 
    ntom_flagSimple = True                   # invoke only listbox form 
    ntom_callback=editSQLTable_ntom_simple
    ntom_table1,ntom_key1=None,None          # master table and key
    ntom_table2,ntom_key2=None,None          # slave table and key
    ntom_tablenm,ntom_tablenm_cols=None,None # in order primary key, key name of table 1, key name of table 2 
    link_ntom_listbox_cfg=None               # multi-choice listbox config
    ntom_copy_cascade = None                 # define which columns of the mn-table to cascade-copy     

##################################################################################################################################################################
# I.6: RefNtoM Non-Simple Listbox/Textframe 
##################################################################################################################################################################
class ntm_nonsimple(SupySQLntomTextFrame_cfg,lb_singlechoice_cfg):
    readonly=True
    title      = None 
    ntom_flagSimple = False               # invoke only listbox form 
    ntom_callback=editSQLTables
    link_ntom_form_cfg=None               # multi-choice listbox config
    ntom_copy_cascade = None              # define which columns of the mn-table to cascade-copy     


##################################################################################################################################################################
##################################################################################################################################################################
#        
# Part II: Dialog definitions of Basic Isolated Tables  
#        
##################################################################################################################################################################
##################################################################################################################################################################

##################################################################################################################################################################
# Dialoge für Table 1: Simulation: Simulation instances for simulation program that use certain parameter sets   
##################################################################################################################################################################
class lb_simulation_cfg(lb_singlechoice_cfg):    # Listbox zur Auswahl des Datensatzes
    tables = ['simulation']
    cols   = ['simulation.key', 'simulation.name', 'date_init', 'date_lastmod', 'grade']
    cols_format = ['5d','15s:15','10s:10','10s:10','5d']
    cols_sortbutton = ['ID','Simulation','create date','last modified','grade']
    title      = 'Choose Simulation Instance'
    width=60

class form_simulation_cfg(form_cfg):            # Eingabform
    tables                = ['simulation']
    join_on               = None 
    cols                  = ['simulation.key','simulation.name','simulation.comment','simulation.date_init','simulation.date_lastmod'  ,\
                             'simulation.simsteps_per_frame'   ,'simulation.frames_per_step'               ,'simulation.delay_per_step','simulation.grade']   
    cols_type             = ['str'           ,'str'            ,'textfield'         ,'str'                 ,'str'                      ,\
                             'str'                             ,'str'                                      ,'str'                      ,'str'             ]
    cols_size             = [40              ,40               ,(40,10)             ,10                    ,10                         ,\
                             10                                ,10                                         ,10                         ,10                ]   
    cols_ref              = None 
    cols_readonly         = [1               ,0                ,0                   ,0                     ,0                          ,\
                             0                                 ,0                                          ,0                          ,0                 ]
    cols_label            = ['ID'            ,'Simulation'     ,'Comment'           ,'Date (creation)'     ,'Date (last mod)'          ,\
                             'Simsteps/Frame'                  ,'Frames/Step'                              ,'Delay/Step'               ,'Grade'           ] 
    cols_label_pos        = None
    cols_helptext         = ['ID of Simulation (integer)', 'Name of the simulation', 'comment on the simulation', 'date of creation of this simulation file', 'date of last modification',\
                             'Number of Simulation Steps per Display Frame','Number of Display Frames per IVisit call to step()','Delay [msec] per IVisit call to step()',\
                             'Grade of the Simulation, that is, an evaluation of the performance of the simulation program with the associated set of parameters']
    select_cols           = ['simulation.key']
    select_vals           = [0]                                      # default value seems irrelevant here (overwritten by tableditor...)
    update_tables = ['simulation']                                   # tables to be updated after form submission
    update_tables_pkeys = [['simulation.key']]                       # for each table (to be updated) a list of primary keys 
    update_tables_cols  = [['simulation.name','simulation.comment','simulation.date_init','simulation.date_lastmod',\
                            'simulation.simsteps_per_frame','simulation.frames_per_step','simulation.delay_per_step','simulation.grade']] # for each table (to be updated) a list of columns to be updated
    title      = 'Edit Simulation Instance ...'                   # is used only for askSupySQLForm

class tbed_simulation_cfg(tbed_config):        # Tabellenbearbeitung
    table_cfg          = tb_simulation_cfg     # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly     = [0]                   # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    cfg_choiceLB       = lb_simulation_cfg     # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm        = form_simulation_cfg   # configuration for record form (typically a SupySQLForm_config)

##################################################################################################################################################################
# Dialoge für Table 2: Parameter: Parameter of the simulation program that may be influenced by IVisit      
##################################################################################################################################################################
class lb_parameter_cfg(lb_singlechoice_cfg):    # Listbox zur Auswahl des Datensatzes
    tables = ['parameter']
    cols   = ['parameter.key', 'parameter.name']
    cols_format = ['5d','35s:35']
    cols_sortbutton = ['ID','Parameter']
    title      = 'Choose Parameter'

class form_parameter_cfg(form_cfg):            # Eingabform
    tables                = ['parameter']
    join_on               = None 
    cols                  = ['parameter.key','parameter.name','parameter.comment','parameter.type'      ,'parameter.range','parameter.listidx'   ]   
    cols_type             = ['str'          ,'str'           , 'str'             ,'optionlist'          ,'str'            ,'str'                 ]
    cols_size             = [40              ,40             ,(40,10)            ,40                    ,40               ,10                    ]   
    cols_ref              = [None            ,None           ,None               ,['int','float','text'],None             ,None                  ]
    cols_readonly         = [1               ,0              ,0                  ,0                     ,0                ,0                     ]
    cols_label            = ['ID'            ,'Parameter'    ,'Comment'          ,'Parameter Type'      ,'Parameter range','Parameter List Index']  
    cols_label_pos        = None
    cols_helptext         = ['ID of Parameter (integer)', 'Name of the parameter', 'Comment on the parameter', '(Element) Type of parameter (either integer, float or text)', \
                             '(Element) Parameter range: either [min,max] or [val1,val2,...] or [] if none',\
                             'Parameter Index if only a single value of a parameter list should be controlled (choose -1 if the whole list should be controlled)']
    select_cols           = ['parameter.key']
    select_vals           = [0]                                # default value seems irrelevant here (overwritten by tableditor...)
    update_tables = ['parameter']                                   # tables to be updated after form submission
    update_tables_pkeys = [['parameter.key']]                       # for each table (to be updated) a list of primary keys 
    update_tables_cols  = [['parameter.name','parameter.comment','parameter.type','parameter.range','parameter.listidx']] # for each table (to be updated) a list of columns to be updated
    title      = 'Edit Parameter Instance ...'                    # is used only for askSupySQLForm

class tbed_parameter_cfg(tbed_config):        # Tabellenbearbeitung
    table_cfg          = tb_parameter_cfg     # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly     = [] #[0]                   # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    cfg_choiceLB       = lb_parameter_cfg     # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm        = form_parameter_cfg   # configuration for record form (typically a SupySQLForm_config)

##################################################################################################################################################################
# Dialoge für Table 3: DataArray: DataArray of the simulation program that may be influenced by IVisit      
##################################################################################################################################################################
class lb_dataarray_cfg(lb_singlechoice_cfg):    # Listbox zur Auswahl des Datensatzes
    tables = ['dataarray']
    cols   = ['dataarray.key', 'dataarray.name','dataarray.type']
    cols_format = ['5d','35s:35','20s:20']
    cols_sortbutton = ['ID','Data array','Type']
    title      = 'Choose Data-Array'

class form_dataarray_cfg(form_cfg):            # Eingabform
    tables                = ['dataarray']
    join_on               = None 
    cols                  = ['dataarray.key','dataarray.name','dataarray.comment','dataarray.type'               ,'dataarray.range']   
    cols_type             = ['str'          ,'str'           , 'str'             ,'optionlist'                   ,'str'            ]
    cols_size             = [40              ,40             ,(40,10)            ,40                             ,40               ]   
    cols_ref              = [None            ,None           ,None               ,['int','float','binary','text'],None             ]
    cols_readonly         = [1               ,0              ,0                  ,0                              ,0                ]
    cols_label            = ['ID'            ,'Name of data' ,'Comment'          ,'Data type'                    ,'Data range'     ]  
    cols_label_pos        = None
    cols_helptext         = ['ID of Data array (integer)', 'Name of the data variable in simulation program', 'Comment on the data variable', 'Type of data (either integer, float, binary or text)', \
                             'Data range in form [min,max]; (relevant only for numeric image-type data)']
    select_cols           = ['dataarray.key']
    select_vals           = [0]                                     # default value seems irrelevant here (overwritten by tableditor...)
    update_tables = ['dataarray']                                   # tables to be updated after form submission
    update_tables_pkeys = [['dataarray.key']]                       # for each table (to be updated) a list of primary keys 
    update_tables_cols  = [['dataarray.name','dataarray.comment','dataarray.type','dataarray.range']] # for each table (to be updated) a list of columns to be updated
    title      = 'Edit Data-Array Instance ...'                    # is used only for askSupySQLForm

class tbed_dataarray_cfg(tbed_config):        # Tabellenbearbeitung
    table_cfg          = tb_dataarray_cfg     # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly     = [] #[0]                   # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    cfg_choiceLB       = lb_dataarray_cfg     # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm        = form_dataarray_cfg   # configuration for record form (typically a SupySQLForm_config)

##################################################################################################################################################################
# Dialoge for Table 4: ParameterWidget: Widget class for displaying and modifying parameters    
##################################################################################################################################################################

class lb_parameterwidget_cfg(lb_singlechoice_cfg):
    tables = ['parameterwidget','simulation','parameter']
    join_on = ['parameterwidget.key_simulation=simulation.key', 'parameterwidget.key_parameter=parameter.key']
    cols   = ['parameterwidget.key','parameterwidget.name','simulation.key','simulation.name','simulation.grade','parameter.key','parameter.name']
    cols_format = ['5d','20s:20','5d','20s:20','5d','5d','20s:20']
    cols_sortbutton = ['ID','NAME','ID-SIM','SIM-NAME','SIM-GRADE','ID-PAR','PAR-NAME']
    width = 80
    title      = 'Choose a ParameterWidget'

class form_parameterwidget_cfg(form_cfg):
    tables                = ['parameterwidget']
    join_on               = None
    cols                  = ['parameterwidget.key'    ,'parameterwidget.key_simulation'      ,'parameterwidget.key_parameter','parameterwidget.name'     ,\
                             'parameterwidget.comment','parameterwidget.type'                ,'parameterwidget.range'        ,'parameterwidget.items'    ,\
                             'parameterwidget.size'   ,'parameterwidget.pos'                 ,'parameterwidget.value']
    cols_type             = ['str'                    ,'ref'                                 ,'ref'                          ,'str'                      ,\
                             'textfield'              ,'optionlist'                          ,'str'                          ,'str'                      ,\
                             'str'                    ,'str'                                 ,'str'                  ] 
    cols_ref              = [None                     ,lb_simulation_cfg                     ,lb_parameter_cfg               ,None                       ,\
                             None                     ,['slider','dictslider','listselection','checkbox','radiobutton','button','textfield'],None                           ,None                       ,\
                             None                     ,None                                  ,None                   ]
    cols_readonly         = [1                        ,0                                     ,0                              ,0                          ,\
                             0                        ,0                                     ,0                              ,0                          ,\
                             0                        ,0                                     ,0                      ]
    cols_label            = ['ID'                     ,'Simulation-ID'                       ,'Parameter-ID'                 ,'Par-Widget-Name'          ,\
                             'Par-Widget-Comment'     ,'Par-Widget-Type'                     ,'Par-Widget-Range'             ,'Par-Widget-Items'         ,\
                             'Par-Widget-Size'        ,'Par-Widget-Position'                 ,'Value'                ] 
    cols_label_pos        = None
    cols_helptext         = ['ID (integer)', 'ID of Simulation Instance', 'ID of Parameter', 'Name of Parameter-Widget', 'Comment on Parameter-Widget', \
                             'Type of Parameter-Widget (either Slider, ListSelection, Checkbox, Radiobutton, Button, or TextField)', \
                             'Range of Parameter-Widget in Format [min, max, nTicks, scale]\n The list fiels nTicks and scale correspond to sliders number of ticks and resolution',\
                             'List of Items for a list selection widget in format [item1,item2,...]', 'Size of Parameter-Widget in Format [sz_x,sz_y]', \
                             'Position of Parameter-Widget in Format [x,y]','Value of Parameter-Widget (as Text)'] 
    #select_cols           = []
    #select_vals           = []                               # default value seems irrelevant here (overwritten by tableditor...)
    select_cols           = ['parameterwidget.key']
    select_vals           = [0]                               # default value seems irrelevant here (overwritten by tableditor...)
    update_tables = ['parameterwidget']                       # tables to be updated after form submission
    update_tables_pkeys = [['parameterwidget.key']]           # for each table (to be updated) a list of primary keys 
    update_tables_cols  = [['parameterwidget.key_simulation','parameterwidget.key_parameter','parameterwidget.name','parameterwidget.comment',\
                            'parameterwidget.type','parameterwidget.range','parameterwidget.items','parameterwidget.size','parameterwidget.pos',\
                            'parameterwidget.value']] # list of columns to be updated
    title      = 'Edit Parameter-Widget...'                   # is used only for askSupySQLForm

class tbed_parameterwidget_cfg(tbed_config):
    table_cfg          = tb_parameterwidget_cfg          # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly     = []                              # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    cfg_choiceLB       = lb_parameterwidget_cfg          # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm        = form_parameterwidget_cfg        # configuration for record form (typically a SupySQLForm_config)
    
##################################################################################################################################################################
# Dialoge for Table 5: DataWidget: Widget class for displaying data        
##################################################################################################################################################################

class lb_datawidget_cfg(lb_singlechoice_cfg):
    tables = ['datawidget','simulation','dataarray']
    join_on = ['datawidget.key_simulation=simulation.key', 'datawidget.key_dataarray=dataarray.key']
    cols   = ['datawidget.key','datawidget.name','simulation.key','simulation.name','dataarray.key','dataarray.name']
    cols_format = ['5d','20s:20','5d','20s:20','5d','20s:20']
    cols_sortbutton = ['ID','NAME','ID-SIM','SIM-NAME','ID-DATA','DATA-NAME']
    width = 80
    title      = 'Choose a Data-Widget'

class form_datawidget_cfg(form_cfg):
    tables                = ['datawidget']
    join_on               = None
    cols                  = ['datawidget.key'  ,'datawidget.key_simulation','datawidget.key_dataarray',                                                              \
                             'datawidget.name' ,'datawidget.comment'       ,'datawidget.type'         ,'datawidget.range' ,'datawidget.pos'      ,'datawidget.scale' ]
    cols_type             = ['str'             ,'ref'                      ,'ref'                     ,                                                              \
                             'str'             ,'textfield'                ,'optionlist'              ,'str'              ,'str'                 ,'str'              ] 
    cols_ref              = [None              ,lb_simulation_cfg          ,lb_dataarray_cfg          ,                                                              \
                             None              ,None                       ,['image','textfield']     ,None               ,None                  ,None               ] 
    cols_readonly         = [1                 ,0                          ,0                         ,                                                              \
                             0                 ,0                          ,0                         ,0                  ,0                     ,0                  ]
    cols_label            = ['ID'              ,'Simulation-ID'            ,'Data-Array-ID'           ,                                                              \
                             'Data-Widget-Name','Data-Widget-Comment'      ,'Data-Widget-Type'        ,'Data-Widget-Range','Data-Widget-Position','Data-Widget-Scale']
    cols_label_pos        = None
    cols_helptext         = ['ID (integer)', 'ID of Simulation Instance', 'ID of Data-Array', 
                             'Name of DataArray-Widget', 'Comment on DataArray-Widget', 'Type of DataArray-Widget (either Image or TextField)', \
                             'Range of DataArray-Widget in Format [min, max] (applies only for image-type data)', 'Position of DataArray-Widget in Format [x,y]', \
                             'Scale used for Dataarray-Widget (applies only for images), e.g., scale=2 means that an image displayed in a widget has double the size of the original image.'+
                             'For Textfields scale<1 means that the widget name is _not_ displayed!']
    select_cols           = ['datawidget.key']
    select_vals           = [0]                               # default value seems irrelevant here (overwritten by tableditor...)
    update_tables = ['datawidget']                       # tables to be updated after form submission
    update_tables_pkeys = [['datawidget.key']]           # for each table (to be updated) a list of primary keys 
    update_tables_cols  = [['datawidget.key_simulation','datawidget.key_dataarray','datawidget.name','datawidget.comment',\
                            'datawidget.type','datawidget.range','datawidget.pos','datawidget.scale']] # list of columns to be updated
    title      = 'Edit Data-Array-Widget...'                      # is used only for askSupySQLForm

class tbed_datawidget_cfg(tbed_config):
    table_cfg          = tb_datawidget_cfg          # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly     = []                              # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    cfg_choiceLB       = lb_datawidget_cfg          # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm        = form_datawidget_cfg        # configuration for record form (typically a SupySQLForm_config)
    

##################################################################################################################################################################
# Dialoge for Table 6: CommentWidget: Widget class for displaying comments        
##################################################################################################################################################################

class lb_commentwidget_cfg(lb_singlechoice_cfg):
    tables = ['commentwidget','simulation']
    join_on = ['commentwidget.key_simulation=simulation.key']
    cols   = ['commentwidget.key','commentwidget.name','simulation.key','simulation.name']
    cols_format = ['5d','20s:20','5d','20s:20']
    cols_sortbutton = ['ID','NAME','ID-SIM','SIM-NAME']
    width = 80
    title      = 'Choose a Comment-Widget'

class form_commentwidget_cfg(form_cfg):
    tables                = ['commentwidget']
    join_on               = None
    cols                  = ['commentwidget.key'     ,'commentwidget.key_simulation','commentwidget.name'     ,'commentwidget.comment','commentwidget.type'           ,'commentwidget.fontname',\
                             'commentwidget.fontsize','commentwidget.fontstyle'     ,'commentwidget.fontcolor','commentwidget.bgcolor','commentwidget.flagDisplayName','commentwidget.pos']
    cols_type             = ['str'                   ,'ref'                         ,'str'                    ,'textfield'            ,'optionlist'                   ,'str'                   ,\
                             'str'                   ,'str'                         ,'str'                    ,'str'                  ,'str'                          ,'str'              ] 
    cols_ref              = [None                    ,lb_simulation_cfg             ,None                     ,None                   ,['textfield']                  ,None                    ,\
                             None                    ,None                          ,None                     ,None                   ,None                           ,None               ] 
    cols_readonly         = [1                       ,0                             ,0                        ,0                      ,0                              ,0                       ,\
                             0                       ,0                             ,0                        ,0                      ,0                              ,0                  ]
    cols_label            = ['ID'                    ,'Simulation-ID'               ,'Widget-Name'            ,'Comment'              ,'Type'                         ,'Font name'             ,\
                             'Font size'             ,'Font style'                  ,'Font color'             ,'Background color'     ,'flagDisplayName'              ,'Widget Position'  ]
    cols_label_pos        = None
    cols_helptext         = ['ID (integer)', 'ID of Simulation Instance', 'Name of Comment-Widget', 'Comment', 'Type of Comment-Widget (either "textfield" or ???)', \
                             'Font name (e.g., Arial)', 'Font size (e.g., 10)', 'Font style (e.g., normal)', 'Font color (e.g., black)','Background color (e.g., gray)', \
                             'Flag indicating whether widget name is displayed (1) or not (0)','Position of DataArray-Widget in Format [x,y]']
    select_cols           = ['commentwidget.key']
    select_vals           = [0]                          # default value seems irrelevant here (overwritten by tableditor...)
    update_tables = ['commentwidget']                       # tables to be updated after form submission
    update_tables_pkeys = [['commentwidget.key']]           # for each table (to be updated) a list of primary keys 
    update_tables_cols  = [['commentwidget.key_simulation','commentwidget.name' ,'commentwidget.comment'       ,'commentwidget.type'         ,\
                            'commentwidget.fontname' ,'commentwidget.fontsize' ,'commentwidget.fontstyle' ,'commentwidget.fontcolor' ,'commentwidget.bgcolor',\
                            'commentwidget.flagDisplayName','commentwidget.pos']] # list of columns to be updated
    title      = 'Edit Comment-Widget...'                      # is used only for askSupySQLForm

class tbed_commentwidget_cfg(tbed_config):
    table_cfg          = tb_commentwidget_cfg          # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly     = []                              # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    cfg_choiceLB       = lb_commentwidget_cfg          # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm        = form_commentwidget_cfg        # configuration for record form (typically a SupySQLForm_config)
    

##################################################################################################################################################################
##################################################################################################################################################################
#        
# Part III: Hub Dialogs (according to expected Workflow)  
#        
##################################################################################################################################################################
##################################################################################################################################################################

##################################################################################################################################################################
# HUB-I Dialog: Erfassung von Simulationen    
##################################################################################################################################################################

#non-simple
class ntom_simulation_parameterwidget_cfg(ntm_nonsimple):    # n-to-m-listbox (!!!!!)
    tables  = ['simulation','parameterwidget','parameter']
    join_on = ['parameterwidget.key_simulation=simulation.key','parameterwidget.key_parameter=parameter.key']
    cols   = ['parameterwidget.key','parameterwidget.name','parameterwidget.type','parameter.name', 'parameterwidget.value']
    cols_format = ['5d','15s:15','10s:10','15s:15','10s:10']
    cols_sortbutton = ['ID','NAME','TYPE','PAR-NAME','VALUE']
    width = 90
    link_ntom_form_cfg=tbed_parameterwidget_cfg
    ntom_copy_cascade = [('simulation','key'), ('parameterwidget','key','key_simulation',['key_parameter','name','comment','type','range','items','size','pos','value'])]  #None    # [(table1,key1),(tablenm,keynm,foreignkey,[list of columns to be copied])]

class ntom_simulation_datawidget_cfg(ntm_nonsimple):    # n-to-m-listbox (!!!!!)
    tables  = ['simulation','datawidget','dataarray']
    join_on = ['datawidget.key_simulation=simulation.key','datawidget.key_dataarray=dataarray.key']
    cols   = ['datawidget.key', 'datawidget.name', 'datawidget.type','dataarray.name', 'dataarray.type']
    cols_format = ['5d','15s:15','10s:10','15s:15','10s:10']
    cols_sortbutton = ['ID','NAME','TYPE','DATA-NAME','DATA-TYPE']
    width = 90
    link_ntom_form_cfg=tbed_datawidget_cfg
    ntom_copy_cascade = [('simulation','key'), ('datawidget','key','key_simulation',['key_dataarray','name','comment','type','range','pos','scale'])]  #None    # [(table1,key1),(tablenm,keynm,foreignkey,[list of columns to be copied])]

class ntom_simulation_commentwidget_cfg(ntm_nonsimple):    # n-to-m-listbox (!!!!!)
    tables  = ['simulation','commentwidget']
    join_on = ['commentwidget.key_simulation=simulation.key']
    cols   = ['commentwidget.key', 'commentwidget.name', 'commentwidget.type']
    cols_format = ['5d','20s:20','20s:20']
    cols_sortbutton = ['ID','NAME','TYPE']
    width = 60
    link_ntom_form_cfg=tbed_commentwidget_cfg
    ntom_copy_cascade = [('simulation','key'), ('commentwidget','key','key_simulation',['name','comment','type','fontname','fontsize','fontstyle','fontcolor','bgcolor','flagDisplayName','pos'])]  # [(table1,key1),(tablenm,keynm,foreignkey,[list of columns to be copied])]

#form...
class formHUB_simulation_cfg(form_cfg):
    tables                = ['simulation']
    join_on               = None 
    cols                  = ['simulation.key'            ,'simulation.name'          ,'simulation.comment','simulation.date_init'             ,'simulation.date_lastmod'     ,'simulation.simsteps_per_frame' ,\
                             'simulation.frames_per_step','simulation.delay_per_step','simulation.grade'  ,None                               ,None                          ,None                             ] 
    cols_type             = ['str'                       ,'str'                      ,'textfield'         ,'str'                              ,'str'                         ,'str'                           ,\
                             'str'                       ,'str'                      ,'str'               ,'ref_ntom'                         ,'ref_ntom'                    ,'ref_ntom'                       ] 
    cols_ref              = [None                        ,None                       ,None                ,None                               ,None                          ,None                            ,\
                             None                        ,None                       ,None                ,ntom_simulation_parameterwidget_cfg,ntom_simulation_datawidget_cfg,ntom_simulation_commentwidget_cfg]
    cols_readonly         = [1                           ,0                          ,0                   ,0                                  ,0                             ,0                               ,\
                             0                           ,0                          ,0                   ,0                                  ,0                             ,0                                ] 
    cols_label            = ['ID'                        ,'Simulation'               ,'Comment'           ,'Date (creation)'                  ,'Date (last mod)'             ,'Simsteps/Frame'                ,\
                             'Frames/Step'               ,'Delay/Step'               ,'Grade'             ,'Parameter-Widgets'                ,'Data-Widgets'                ,'Comment-Widgets'                ] 
    cols_size             = [40                          ,40                         ,(40,7)              ,10                                 ,10                            ,10                              ,\
                             10                          ,10                         ,10                  ,(70,10)                            ,(70,5)                        ,(70,5)]     # may also be list
    cols_helptext         = ['ID of Simulation (integer)'       , 'Name of the simulation'    ,'Comment on the simulation','Date of creation of this simulation file', 'Date of last modification', \
                             'Number of Simulation Steps per Display Frame','Number of Display Frames per IVisit call to step()','Delay [msec] per IVisit call to step()', \
                             'Grade of the simulation, that is, a performance measure of the simulation program with the associated parameter values (choose by hand)', \
                             'Parameter widgets for this simulation', 'Data widgets for this simulation','Comment widgets for this simulation']
    #cols2copy             = ['picture','key_sex','key_religion']   # if None all columns can be copied (e.g., by Tableeditor.on_lb_copy...); otherwise list of columns (of main table!) to be copied
    select_cols           = ['simulation.key']
    select_vals           = [0]                                   # default value seems irrelevant here (overwritten by tableditor...)
    update_tables = ['simulation']                       # tables to be updated after form submission
    update_tables_pkeys = [['simulation.key']]           # for each table (to be updated) a list of primary keys 
    update_tables_cols  = [['simulation.name','simulation.comment','simulation.date_init','simulation.date_lastmod',\
                            'simulation.simsteps_per_frame','simulation.frames_per_step','simulation.delay_per_step','simulation.grade']] # for each table (to be updated) a list of columns to be updated
    title      = 'Edit Simulation and associated parameter, data, and comment widgets...'                      # is used only for askSupySQLForm

#tableeditor...
class tbedHUB_simulation_cfg(tbed_config):
    table_cfg          = tb_simulation_cfg          # table to be edited, first column is assumed to be primary key of type INTEGER!!!
    pkeys_readonly     = [0]                        # records with primary key in this list are considered as READ_ONLY (e.g., for default records)
    cfg_choiceLB       = lb_simulation_cfg          # configuration for choice listbox (typically a SQLListbox_config)
    cfg_recForm        = formHUB_simulation_cfg     # configuration for record form (typically a SupySQLForm_config)








##################################################################################################################################################################
##################################################################################################################################################################
# Module test
##################################################################################################################################################################
##################################################################################################################################################################

if __name__ == '__main__':
    print("\nModule test of ivisit.defdb.py")
    print("------------------------------------\n") 
    db = sqldatabase(db_ivisit_cfg)
    #db.print_database(1)
    r=None 
    res = editSQLTables(r,db,"Edit DB-Table Simulation",tbed_simulation_cfg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
    
# *******************************************
# *******************************************
# Help text
# *******************************************
# *******************************************

# *******************************************
# (i) README.txt
# *******************************************

readme_text_ivisit =\
"""---------------------------------------
---------------------------------------
README.txt (currently only in German):
---------------------------------------
---------------------------------------

INSTALLATION FÜR LINUX UND UBUNTU:
----------------------------------------
----------------------------------------

Installation mit pip: pip3 install ivisit
bzw. mit allen abhängigen Paketen: pip3 install ivisit supylib pillow numpy

Oder alte Methode vom tar.gz-Archiv: 

1) IVisit.0.60.tar.gz in ein beliebiges Verzeichnis (z.B. Ihrer Python-Projekte) kopieren und mit
      tar xfz IVisit.0.60.tar.gz 
   entpacken.

2) In Ihrer .bashrc eine Zeile für den PYTHONPATH auf das Verzeichnis IVisit.0.60 einfügen (damit
   die Python-Module von IVisit überall sichtbar sind), z.B.
      export PYTHONPATH=/home/login-name/projects/IVisit.0.60:$PYTHONPATH

Alternative: Statt obiger zwei Zeilen in der .bashrc reicht auch das Einfügen der folgenden Zeile:
source /home/login-name/projects/IVisit/python/BashSrc

Bemerkung: Wenn mehrere Versionen von IVisit und/oder ReLabelEd parallel verwendet werden, dann können
           Sie durch den obigen source-Befehl (in einer Shell ausgeführt) diese Version aktivieren
	   (es wird dadurch PATH und PYTHONPATH geupdated, sodass sie Verzeichnisse der ausgewählten
	   Version ganz vorne in den Suchpfaden erscheinen und deshalb zuerst ausgeführt werden).

3) Starten/Test/Demo: Neue Bash öffnen (damit neuer PYTHONPATH wirksam wird) und in
   das Verzeichnis Ivisit/demos gehen. Dann den Befehl
      python demo1_gaussfilter.py demo1_gaussfilter.db&
   eingeben („&“ setzt einen eigenen Prozess ab, so dass man in der Shell weiterhin Befehle eingeben kann).
   Es sollte nun das Ivisit-Fenster starten in dem ein Demoprogramm aus der Bildverarbeitung ausgeführt
   wird (Gauß-Filter). Sie können die „Simulation“ starten indem Sie in der Button-Leiste links unten
   „Run“ drücken. In jedem Simulationsschritt werden die Parameter von den GUI-Elementen (hier: Bild-Index,
   Kernel-Radius, Kernel-Sigma) gelesen und dann der Simulationszustand geupdated (hier: das gefilterte
   Bild „Filter-Output“). Spielen Sie mit den Sliders! Oben rechts sehen Sie die Anzahl der
   Simulationsschritte. Wenn Sie in der Button-Leiste links unten auf „Stop“ drücken hält die Simulation an.
   Mit „Cont“ können Sie fortsetzen. Mit „Run“ neu starten. Mit „Init“ die Simulation initialisieren.
   Mit „Save“ speichern Sie die aktuellen Parametereinstellungen. Mit „Quit“ unten rechts können Sie die IVisit beenden.
   Ähnlich können Sie die zweite etwas komplexere Simulation starten:
      python demo2_DoG_filter.py demo2_DoG_filter.db&

4) Wichtigste Bedienelemente für eine gegebene IVisit-Simulation (wie z.B. die Demo-Simulationen)

4.1) Die Action-Button-Leiste links unten:
Save: Speichert die aktuellen Einstellungen der Parameter und Visulisierungs-Widgets
Init: Initialisiert die Simulation durch Aufrufen der init() Methode
Step: Es wird ein Simulationsschritt durch Aufrufen der step() Methode durchgeführt
Run: Es wird init() und danach wiederholt die step() Methode aufgerufen. Zwischen zwei Aufrufen der step() Methode wird 
     ein bestimmtes Zeitintervall pausiert. Die Länge des Zeitintervalls können Sie im Menüpunkt Databases/Simulations 
     durch das Attribut delay/step einstellen.
Stop: Die durch Run gestartete automatische Simulation wird gestoppt.
Cont: Fortsetzen der Simulation (ohne Initialisieren)
Simulations: Aufruf des Menüpunkts Databases/Simulations, z.B. zur Auswahl einer bestimmten zuvor gespeicherten
             Simulations- bzw. Parameterkonfiguration.
Parse: Die im Python-Simulationsskript definierten GUI-Elemente werden gescannt und in die Simulation eingefügt 
       (falls sie nicht schon vorhanden sind). Dies ist meist beim ersten Ausführen einer Simulation nötig, damit man
       nicht alle GUI-Elemente von Hand über Menüpunkt Databases einfügen muss. 

4.2) Die Action Button-Leiste rechts unten:
Quit: Beenden von IVisit
Help: Hilfe

4.3) Der schwarz unterlegte Info-Balken unter der Menüleiste <File-Name>::<ID>::<Simulations-Name> und <Steps>
File-Name: Der File-Name ganz links zeigt das Aktuell verwendete Daten-File (Endung *.db) an mit der die Simulation
           läuft. In diesem File werden die Parametereinstellungen zu den jeweiligen Simulations-Konfigurationen
           gespeichert.
Simulations-ID/Name: Die ID und der Name der aktuell verwendeten Simulations-Konfiguration. Sie können diese über Action-
                     Button Simulations bzw. Menüpunkt Databases/Simulations auswählen/ändern/neu erstellen.
Steps: Gibt an wieviele Simulationsschritte schon durchgeführt wurden.

4.4) Menüleiste:
File: Laden/Speichern eines Daten-Files zu einer Simulation (mit Endung *.db)
HUBS: Wichtigste Action-Dialoge:
HUBS/Simulations: Auswahl/Neuerstellung einer Simulations-Konfiguration: In jedem Daten-File können Sie beliebige
                  Simulations-Konfigurationen zu einer Simulation speichern, typischerweise um interessante Parameter-
                  Werte zu speichern, während man weiter nach noch besseren sucht...
HUBS/Edit-Parameter-Widgets: Verwaltung der Parameter-Widgets (siehe auch Databases/Parameter-Widgets). Parameter-Widgets
                  sind die GUI-Input-Elemente, die von der Simulation gelesen werden, und mit denen Sie typischerweise
                  System-Parameter der Simulation beeinflussen. Parameter-Widgets sind z.B. Sliders, List-Selections und
                  Textfelder. Jeder Parameter-Widget ist mit einem Parameter verbunden (siehe auch Databases/Parameter),
                  der wiederum mit einer Variablen aus dem Python-Simulations-Skript verbunden ist.
HUBS/Edit-Data-Widgets: Verwaltung der Daten-Widgets (siehe auch Databases/Data-Widgets). Data-Widgets sind die Output-
                  Elemente der GUI, mit denen System-Zustände von der Simulation ausgegeben bzw. visualisiert werden
                  können, damit das System-Verhalten beobachtet werden kann. Daten-Widgets sind z.B. Images oder
                  Textfelder. Jeder Daten-Widget ist mit einem Data-Array verbunden (siehe auch Databases/Data-Array),
                  das wiederum mit einer Variablen aus dem Python-Simulations-Skript verbunden ist.
Databases: Hier können Sie die einzelnen Tabellen des Data-Files (mit Endung *.db) bearbeiten.
            - Die meisten Menü-Punkte entsprechen den vorigen Menüpunkten
            - Mit dem Menüpunkt Databases/Comment-Widget kann man Kommentare einfügen mit dem man die Abläufe in der
              Simulation beschreiben und erklären kann.      
Simulation: Hier können Sie die Simulations-Funktionen steuern (entspricht den Aktion-Buttons unten links)
Help: Hilfefunktion

4.5) Simulations-Fläche in der Mitte
In der Mitte des IVisit-Fensters werden die Parameter-Widgets (zur Eingabe), die Data-Widgets (zur Ausgabe), und die
Comment-Widgets (zur Erklärung) dargestellt. Sie können frei verschoben werden, wobei die Positionen der Widgets
ebenfalls mit „Save“ gespeichert werden.

5) Erstellen eines eigenen Python-Simulations-Skripts
- Schauen Sie sich am besten den Aufbau der Demo-Simulations-Skripte an (demo1_gaussfilter.py und demo2_DoG_filter.py)
- Jedes Simulations-Skript enthält die folgenden Teile:
    - class SimParameters: Definiert die Simulations-Parameter die Sie über IVisit verändern/einstellen möchten
    - class SimData: Definiert die Simulations-Daten die Sie über IVisit ausgeben/visualisieren möchten
    - class Sim: Definiert die eigentliche Simulation über die Methoden
         - main_init(): Wird einemal beim Start bzw. Konstruktion der Simulation/Simulationsobjekte aufgerufen
         - init(): Wird vor jedem Neustart der Simulation aufgerufen
         - step(): Wird in jedem Simulations-Schritt aufgerufen
      Siehe auch die Basis-Klasse IVisit_Simulation im Modul ivisit.simulation.py
    - Hauptprogramm („main program“): Dort wird ein Simulations-Objekt vom Typ Sim erzeugt und durch 
      Ivisit_main(sim=sim) die Ivisit-GUI gestartet
- Definition der GUI-Elemente: 
    - Alle GUI-Elemente können im Prinzip in der IVisit-GUI von Hand erzeugt werden (z.B. unter HUBS oder Databases)
    - Es ist aber meist viel bequemer die (wichtigsten) GUI-Elemente schon im Python-Simulations-Skript zu definieren
    - z.B. Siehe in  demo1_gaussfilter.py, Zeilen 10-17 und die folgende nähere Erklärung...
- Mit der folgenden Syntax kann man im Python-Simulations-Skript Ivisit-GUI-Elemente erzeugen:
    - #@IVISIT:SIMULATION & sim_gauss_filter1  
      erzeugt einen Eintrag in der Database Simulations mit Namen sim_gauss_filter1 
    - #@IVISIT:SLIDER     & Bild-Index    & [200,1] & [0,20,5,1] & idx_image & -1 & int & 0 
         - erzeugt einen Slider mit Namen Bild_index mit Breite 200 Pixel (der zweite Wert 1 ist im Prinzip die Höhe, 
           wird aber für Slider ignoriert). 
         - [0,20,5,1] bedeutet, dass der Wertebereich des Sliders von 0 bis 20 ist, dass er 5 Ticks hat, und die
           Skalierung 1 ist (=Abstand zwischen zwei möglichen Werten)
         - idx_image: Mit dem Slider kann man die Python-Variable idx_image verändern
         - -1: Parameter List index (Index, falls idx_image eine Liste ist; -1 bedeutet, dass es keine Liste ist)
         - int: Data Type der Python-Variable
         - 0: Initialwert des Parameters
    - #@IVISIT:DICTSLIDER  & Kernel-Parameters  & [200,20,-1,2,10] & dict_par & 0 
      #@IVISIT:DICTSLIDERITEM & First Item & [0,20,5,1] & item1 & int & 3 
      #@IVISIT:DICTSLIDERITEM & Second Item & [0,30,4,2] & item2 & float & 3.5
         - erzeugt einen DICT-SLIDER mit dem man die Items eines Parmeter-Dicts verändern kann (unter Vorauswahl eines Items)
         - Obiges Parameter erzeugt einen DICT-SLIDER mit Namen "Kernel-Parameters", Größen-Spezifikation [200,20,-1,2,10] 
           für Dict-Parameter "dict_par", wobei zu Beginn der 0-te Item-Key ausgewählt wird
         - Zur Größen-Spezifikation: [SliderWidth,columns,rows,DisplayMode,FontSize] (hier [200,20,-1,2,10])
             - DisplayMode bestimmt die Darstellung: 
                 0=Slider und einfaches OptionMenu zur Auswahl des Dict-Items
                 1=Slider und OptionMenu mit Angabe der ParameterWerte
                 2=Slider, einfaches OptionMenu und Text-Feld zur Ausgabe+manuellen Eingabe (press "Set") der Parameterwerte 
             - SliderWidth=Breite Slider (hier 200 Pixel)
             - Columns=Breite des Text-Feldes (in Zeichen) bei DisplayMode=2
             - Rows   =Höhe des Text-Feldes (falls -1, dann Rows=Anzahl Items) bei DisplayMode=2
             - FontSize=Schriftgröße bei Display-Mode=2                                 
         - Spezifikation der DICTSLIDERITEMS: Für jedes Dict-Item braucht man eine eigene Zeile mit DICTSLIDERITEM
             - obiges Beispiel hat zwei DICTSLIDERITEMs für zwei Dict-Items dict_par['item1'] und dict_par['item2']
             - Format:  #@IVISIT:DICTSLIDERITEM & <ItemName> & <RangeList> & <Item> & <type> & <value>
                 - <ItemName> wird in OptionMenu/Text-Feld angezeigt
                 - <RangeList>=[min,max,nticks,scale] definiert Wertebereich des Sliders wie bei SLIDER
                 - <ITEM>=Key des Items im Prameter Dict
                 - <type>=Typ des Items (entweder "int" oder "float")
                 - <value>=initialer Parameter-Wert des Items
         - Zu einem DICTSLIDER können beliebig viele (aber mindest ein) DICTSLIDERITEMS hinzugefügt werden
    - #@IVISIT:TEXT_IN & txt-dummy     & [20,5] & dummytextpar & -1 & InitialText
         - erzeugt ein Text-Eingabefeld mit Namen „txt-dummy“ der Größe 20x5 mit dem man den Wert der Python-String-
           Variablen dummytextpar belegen kann. 
         - -1 kann man wieder ignorieren (Listen-Index, falls es sich bei der Python-Variable um Liste handelt)
         - „InitialText“ ist wieder Initialwert
    - #@IVISIT:LISTSEL & lst-dummylist & [20,5] & [A,BBBBB,CCC,DDDDD] & dummytextpar & -1 & string & A
         - erzeugt eine List-Selektion mit Namen „lst-dummylist“ der Größe 20x5
         - man kann zwischen den Einträgen "A","BBBBB","CCC","DDDDD" auswählen
         - Der ausgewählte String wird in der Python-Variablen dummytextpar gespeicherte
         - -1 ist wieder Listen-Index (-1 bedeutet keine Liste)
         - string und „A“ sind Datentyp und Initialwert
	 - verwenden Sie in obiger Definitionszeile keine(!) Anführungszeichen „A“ oder ‚A‘ !!
    - #@IVISIT:CHECKBOX & chbx-dummy & [A,BBBBB,CCC,DDDDD] & dummystr & 0110
         - erzeugt eine Checkbox mit Namen „chbx-dummy“ 
         - man kann jeweils die Einträge "A","BBBBB","CCC","DDDDD" auswählen (binär, Kreuz=‘1‘ oder Leer=‘0‘)
         - Der ausgewählte String (z.B: „0110“, eine Ziffer pro Eintrag) wird im Python-String dummystr gespeicherte
         - String ‘0110‘ ist Initial-Wert als String 
	 - verwenden Sie in obiger Definitionszeile keine(!) Anführungszeichen „0110“ oder ‚0110‘ !!
    - #@IVISIT:RADIOBUTTON & rdbt-dummy & [A,BBBBB,CCC,DDDDD] & dummystr & A
         - erzeugt eine Radiobutton-Box mit Namen „rdbt-dummy“ 
         - man kann genau einen der String Einträge "A","BBBBB","CCC","DDDDD" auswählen 
         - Der ausgewählte String wird im Python-String dummystr gespeicherte
         - String ‘A‘ ist Initial-Wert als String 
	 - verwenden Sie in obiger Definitionszeile keine(!) Anführungszeichen für Strings! 
    - #@IVISIT:BUTTON & bt-dummy & [labeltext,buttontext] & dummystr
         - erzeugt einen Button mit Namen „bt-dummy“
         - Darstellung als „labeltext buttontext“, empfohlen ist „< buttontext“
         - mit labeltext/< kann man Button bewegen; buttontext ist zum Anklicken um den Button zu aktivieren
         - Falls Button geklickt wird, wird Variable dummystr=‘1‘ gesetzt (ansonsten ‘0‘)
         - nach Aufruf von step wird dummystr=‘0‘ zurückgesetzt
	 - verwenden Sie in obiger Definitionszeile keine(!) Anführungszeichen für Strings! 
    - #@IVISIT:IMAGE   & Grauwert-Bild     & 1.0    & [0,255]  & im_gray   & int
         - erzeugt eine Bild mit Namen „Grauwert-Bild“, mit Skalierungsfaktor 1.0 und Wertebereich [0,255] vom Typ int
         - In jedem Simulationsschritt (step) wird der Inhalt der Python-Variable im_gray als Bild im IVisit-GUI 
           dargestellt
    - #@IVISIT:TEXT_OUT & txtout-dummy & [20,5] & list_of_options & dummytextout
         - erzeugt Text-Ausgabe-Box mit Namen txtout-dummy der Größe 20x5, wobei in jedem Simulations-Schritt der 
           Inhalt der Python-Variable als String angezeigt wird
         - list_of_options ist entweder 'None' oder ein einzelner String oder eine Liste aus Strings; 
           mögliche Strings sind 'just_left','just_right','just_center' (default center)
"""





import numpy as np
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure

# ************************************************************************
# Module ivisit.matplotlib:
# Functions for easy inclusion of matplotlib plots in IVisit simulations 
# ************************************************************************

def getMatplotlibFigure(**kwargs):
    fig=Figure(**kwargs)
    return fig

def getMatplotlibImage(fig):    # return rgb image of matplotlib figure
    canvas=FigureCanvas(fig)
    canvas.draw()
    image_from_plot = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
    image_from_plot = image_from_plot.reshape(fig.canvas.get_width_height()[::-1] + (3,))
    return image_from_plot

def getDataPos_from_PixelPos(pixel_x,pixel_y, canvas_h, ax=None, transData2Pixel=None):
    """ 
    transform pixel position within figure canvas to data position within matplotlib axis
    see also https://matplotlib.org/2.0.2/users/transforms_tutorial.html
            https://stackoverflow.com/questions/13662525/how-to-get-pixel-coordinates-for-matplotlib-generated-scatterplot
    """
    assert not(ax is None and transData2Pixel is None), "Parameters ax and transData2Pixel must not both be None!"
    if transData2Pixel is None: transData2Pixel=ax.transData.inverted()
    dataxy=transData2Pixel.transform([(pixel_x,canvas_h-pixel_y)])
    data_x=dataxy[0][0]
    data_y=dataxy[0][1]
    return data_x,data_y

def getDataPos_from_canvas(pixel_x,pixel_y,im_widget_canvas,ax=None, transData2Pixel=None):
    """ 
    transform pixel position within figure canvas to data position within matplotlib axis
    same as getDataPos_from_PixelPos, but simpler interface for ImageWidgets
    im_widget_canvas is binding object (i.e., the canvas of the image widget): it is returned by bind function display.bind2Widget(.) 
    Example: 
       (i) Do binding by 
           self.imw_canvas=display.bind2Widget("Trajectory","<Button-1>",self.onPressedB1_trajectory,"imgcanvas")
       (ii) do call to getDataPosImageWidget(.) by 
           data_x,data_y=getDataPos(self.clicked_x,self.clicked_y,self.imw_canvas,self.ax)
    see also https://matplotlib.org/2.0.2/users/transforms_tutorial.html
            https://stackoverflow.com/questions/13662525/how-to-get-pixel-coordinates-for-matplotlib-generated-scatterplot
    """
    return getDataPos_from_PixelPos(pixel_x,pixel_y, im_widget_canvas.winfo_height(), ax, transData2Pixel)
